from __future__ import annotations

from collections import defaultdict
from datetime import datetime
import json
import logging
from sqlite3 import OperationalError
from typing import (
    Any,
    ClassVar,
    Dict,
    Generator,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
)
import warnings

from alembic.ddl.impl import DefaultImpl
import numpy as np
from packaging.version import Version
import pandas as pd
import pydantic
from pydantic import Field
import sqlalchemy as sa
from sqlalchemy import Table
from sqlalchemy import inspect
from sqlalchemy.orm import joinedload
from sqlalchemy.orm import sessionmaker
from sqlalchemy.sql import text as sql_text
from trulens.core.database import base as core_db
from trulens.core.database import exceptions as db_exceptions
from trulens.core.database import migrations as db_migrations
from trulens.core.database import orm as db_orm
from trulens.core.database import utils as db_utils
from trulens.core.database.legacy import migration as legacy_migration
from trulens.core.database.migrations import data as data_migrations
from trulens.core.otel.utils import is_otel_tracing_enabled
from trulens.core.schema import app as app_schema
from trulens.core.schema import base as base_schema
from trulens.core.schema import dataset as dataset_schema
from trulens.core.schema import feedback as feedback_schema
from trulens.core.schema import groundtruth as groundtruth_schema
from trulens.core.schema import record as record_schema
from trulens.core.schema import types as types_schema
from trulens.core.schema.event import Event
from trulens.core.utils import pyschema as pyschema_utils
from trulens.core.utils import python as python_utils
from trulens.core.utils import serial as serial_utils
from trulens.core.utils import text as text_utils
from trulens.otel.semconv.trace import ResourceAttributes
from trulens.otel.semconv.trace import SpanAttributes

logger = logging.getLogger(__name__)


class SnowflakeImpl(DefaultImpl):
    __dialect__ = "snowflake"


class SQLAlchemyDB(core_db.DB):
    """Database implemented using sqlalchemy.

    See abstract class [DB][trulens.core.database.base.DB] for method reference.
    """

    model_config: ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(
        arbitrary_types_allowed=True
    )

    table_prefix: str = core_db.DEFAULT_DATABASE_PREFIX
    """The prefix to use for all table names.

    [DB][trulens.core.database.base.DB] interface requirement.
    """

    engine_params: dict = Field(default_factory=dict)
    """SQLAlchemy-related engine params."""

    session_params: dict = Field(default_factory=dict)
    """SQLAlchemy-related session."""

    engine: Optional[sa.Engine] = None
    """SQLAlchemy engine."""

    session: Optional[sessionmaker] = None
    """SQLAlchemy session(maker)."""

    orm: Type[db_orm.ORM]
    """Container of all the ORM classes for this database.

    This should be set to a subclass of
    [ORM][trulens.core.database.orm.ORM] upon initialization.
    """

    def __str__(self) -> str:
        """Relatively concise identifier string for this instance."""

        if self.engine is None:
            return "SQLAlchemyDB(no engine)"

        return f"SQLAlchemyDB({self.engine.url.database})"

    # for DB's WithIdentString mixin
    def _ident_str(self) -> str:
        """Even more concise identifier string than __str__."""

        if self.engine is None:
            return "(no engine)"

        if self.engine.url.database is None:
            return f"{self.engine.url.drivername} db"

        return self.engine.url.database

    def __init__(
        self,
        redact_keys: bool = core_db.DEFAULT_DATABASE_REDACT_KEYS,
        table_prefix: str = core_db.DEFAULT_DATABASE_PREFIX,
        **kwargs: Dict[str, Any],
    ):
        super().__init__(
            redact_keys=redact_keys,
            table_prefix=table_prefix,
            orm=db_orm.make_orm_for_prefix(table_prefix=table_prefix),
            **kwargs,
        )
        self._reload_engine()
        if db_utils.is_memory_sqlite(self.engine):
            warnings.warn(
                UserWarning(
                    "SQLite in-memory may not be threadsafe. "
                    "See https://www.sqlite.org/threadsafe.html"
                )
            )

    def _reload_engine(self):
        if self.engine is None:
            # Check if the dialect is snowflake and set isolation_level
            snowflake_sqlalchemy_version = None
            try:
                import snowflake.sqlalchemy

                snowflake_sqlalchemy_version = snowflake.sqlalchemy.__version__
            except Exception:
                pass
            if (
                snowflake_sqlalchemy_version
                and Version(snowflake_sqlalchemy_version) >= Version("1.7.2")
                and "url" in self.engine_params
                and "snowflake" in self.engine_params["url"]
            ):
                temp_engine = sa.create_engine(**self.engine_params)
                if temp_engine.dialect.name == "snowflake":
                    self.engine_params.setdefault(
                        "isolation_level", "AUTOCOMMIT"
                    )
                # Dispose of the temporary engine
                temp_engine.dispose()
            self.engine = sa.create_engine(**self.engine_params)
        self.session = sessionmaker(self.engine, **self.session_params)

    @classmethod
    def from_tru_args(
        cls,
        database_url: Optional[str] = None,
        database_engine: Optional[sa.Engine] = None,
        database_redact_keys: Optional[
            bool
        ] = core_db.DEFAULT_DATABASE_REDACT_KEYS,
        database_prefix: Optional[str] = core_db.DEFAULT_DATABASE_PREFIX,
        **kwargs: Dict[str, Any],
    ) -> SQLAlchemyDB:
        """Process database-related configuration provided to the [Tru][trulens.core.session.TruSession] class to
        create a database.

        Emits warnings if appropriate.
        """

        if "table_prefix" not in kwargs:
            kwargs["table_prefix"] = database_prefix

        if "redact_keys" not in kwargs:
            kwargs["redact_keys"] = database_redact_keys

        if database_engine is not None:
            new_db: core_db.DB = SQLAlchemyDB.from_db_engine(
                database_engine, **kwargs
            )
        else:
            if database_url is None:
                database_url = f"sqlite:///{core_db.DEFAULT_DATABASE_FILE}"
            new_db: core_db.DB = SQLAlchemyDB.from_db_url(
                database_url, **kwargs
            )

        print(
            "%s Initialized with db url %s ."
            % (text_utils.UNICODE_SQUID, new_db.engine.url)
        )
        if database_redact_keys:
            print(
                f"{text_utils.UNICODE_LOCK} Secret keys will not be included in the database."
            )
        else:
            print(
                f"{text_utils.UNICODE_STOP} Secret keys may be written to the database. "
                "See the `database_redact_keys` option of `TruSession` to prevent this."
            )

        return new_db

    @classmethod
    def from_db_url(cls, url: str, **kwargs: Dict[str, Any]) -> SQLAlchemyDB:
        """
        Create a database for the given url.

        Args:
            url: The database url. This includes database type.

            kwargs: Additional arguments to pass to the database constructor.

        Returns:
            A database instance.
        """

        # Params needed for https://github.com/truera/trulens/issues/470
        # Params are from
        # https://stackoverflow.com/questions/55457069/how-to-fix-operationalerror-psycopg2-operationalerror-server-closed-the-conn

        default_engine_params = {
            "url": url,
            "pool_size": 10,
            "pool_recycle": 300,
            "pool_pre_ping": True,
        }

        if not db_utils.is_memory_sqlite(url=url):
            # These params cannot be given to memory-based sqlite engine.
            default_engine_params["max_overflow"] = 2
            default_engine_params["pool_use_lifo"] = True

        if "engine_params" in kwargs:
            for k, v in default_engine_params.items():
                if k not in kwargs["engine_params"]:
                    kwargs["engine_params"][k] = v
        else:
            kwargs["engine_params"] = default_engine_params

        return cls(**kwargs)

    @classmethod
    def from_db_engine(
        cls, engine: sa.Engine, **kwargs: Dict[str, Any]
    ) -> SQLAlchemyDB:
        """
        Create a database for the given engine.
        Args:
            engine: The database engine.
            kwargs: Additional arguments to pass to the database constructor.
        Returns:
            A database instance.
        """

        return cls(engine=engine, **kwargs)

    def check_db_revision(self):
        """See
        [DB.check_db_revision][trulens.core.database.base.DB.check_db_revision]."""

        if self.engine is None:
            raise ValueError("Database engine not initialized.")

        db_utils.check_db_revision(self.engine, self.table_prefix)

    def get_db_revision(self) -> Optional[str]:
        if self.engine is None:
            raise ValueError("Database engine not initialized.")
        return db_migrations.get_current_db_revision(
            self.engine, self.table_prefix
        )

    def migrate_database(self, prior_prefix: Optional[str] = None):
        """See [DB.migrate_database][trulens.core.database.base.DB.migrate_database]."""

        if self.engine is None:
            raise ValueError("Database engine not initialized.")

        try:
            # Expect to get the the behind exception.
            db_utils.check_db_revision(
                self.engine, prefix=self.table_prefix, prior_prefix=prior_prefix
            )

            # If we get here, our db revision does not need upgrade.
            logger.warning("Database does not need migration.")

        except db_exceptions.DatabaseVersionException as e:
            if e.reason == db_exceptions.DatabaseVersionException.Reason.BEHIND:
                revisions = db_migrations.DbRevisions.load(self.engine)
                from_version = revisions.current
                ### SCHEMA MIGRATION ###
                if db_utils.is_legacy_sqlite(self.engine):
                    raise RuntimeError(
                        "Migrating legacy sqlite database is no longer supported. "
                        "A database reset is required. This will delete all existing data: "
                        "`TruSession.reset_database()`."
                    ) from e

                else:
                    ## TODO Create backups here. This is not sqlalchemy's strong suit: https://stackoverflow.com/questions/56990946/how-to-backup-up-a-sqlalchmey-database
                    ### We might allow migrate_database to take a backup url (and suggest user to supply if not supplied ala `TruSession().migrate_database(backup_db_url="...")`)
                    ### We might try copy_database as a backup, but it would need to automatically handle clearing the db, and also current implementation requires migrate to run first.
                    ### A valid backup would need to be able to copy an old version, not the newest version
                    db_migrations.upgrade_db(
                        self.engine, revision="head", prefix=self.table_prefix
                    )

                self._reload_engine()  # let sqlalchemy recognize the migrated schema

                ### DATA MIGRATION ###
                data_migrations.data_migrate(self, from_version)
                return

            elif (
                e.reason == db_exceptions.DatabaseVersionException.Reason.AHEAD
            ):
                # Rethrow the ahead message suggesting to upgrade trulens.
                raise e

            elif (
                e.reason
                == db_exceptions.DatabaseVersionException.Reason.RECONFIGURED
            ):
                # Rename table to change prefix.

                prior_prefix = e.prior_prefix

                logger.warning(
                    'Renaming tables from prefix "%s" to "%s".',
                    prior_prefix,
                    self.table_prefix,
                )
                # logger.warning("Please ignore these warnings: \"SAWarning: This declarative base already contains...\"")

                with self.engine.connect() as c:
                    for table_name in ["alembic_version"] + [
                        c._table_base_name
                        for c in self.orm.registry.values()
                        if hasattr(c, "_table_base_name")
                    ]:
                        old_version_table = f"{prior_prefix}{table_name}"
                        new_version_table = f"{self.table_prefix}{table_name}"

                        logger.warning(
                            "  %s -> %s", old_version_table, new_version_table
                        )

                        c.execute(
                            sql_text(
                                """ALTER TABLE %s RENAME TO %s;"""
                                % (old_version_table, new_version_table)
                            )
                        )

            else:
                # TODO: better message here for unhandled cases?
                raise e

    def reset_database(self):
        """See [DB.reset_database][trulens.core.database.base.DB.reset_database]."""

        # meta = MetaData()
        meta = self.orm.metadata

        tables = [
            Table(f"{self.table_prefix}alembic_version", self.orm.metadata)
        ] + [
            c.__table__
            for c in self.orm.registry.values()
            if hasattr(c, "__table__")
        ]
        meta.drop_all(bind=self.engine, tables=tables)

        self.migrate_database()

    def insert_record(
        self, record: record_schema.Record
    ) -> types_schema.RecordID:
        """See [DB.insert_record][trulens.core.database.base.DB.insert_record]."""
        if is_otel_tracing_enabled():
            raise RuntimeError("Not supported with OTel tracing enabled!")
        # TODO: thread safety

        _rec = self.orm.Record.parse(record, redact_keys=self.redact_keys)
        with self.session.begin() as session:
            if (
                session.query(self.orm.Record)
                .filter_by(record_id=record.record_id)
                .first()
            ):
                session.merge(_rec)  # update existing
            else:
                session.merge(_rec)  # add new record # .add was not thread safe

            logger.info(
                "%s added record %s", text_utils.UNICODE_CHECK, _rec.record_id
            )

            return _rec.record_id

    def batch_insert_record(
        self, records: List[record_schema.Record]
    ) -> List[types_schema.RecordID]:
        if is_otel_tracing_enabled():
            raise RuntimeError("Not supported with OTel tracing enabled!")
        """See [DB.batch_insert_record][trulens.core.database.base.DB.batch_insert_record]."""
        with self.session.begin() as session:
            records_list = [
                self.orm.Record.parse(r, redact_keys=self.redact_keys)
                for r in records
            ]
            session.add_all(records_list)
            logger.info(f"{text_utils.UNICODE_CHECK} added record batch")
            # return record ids from orm objects
            return [r.record_id for r in records_list]

    def get_app(
        self, app_id: types_schema.AppID
    ) -> Optional[serial_utils.JSONized]:
        """See [DB.get_app][trulens.core.database.base.DB.get_app]."""

        with self.session.begin() as session:
            if (
                _app := session.query(self.orm.AppDefinition)
                .filter_by(app_id=app_id)
                .first()
            ):
                return json.loads(_app.app_json)

    def update_app_metadata(
        self, app_id: types_schema.AppID, metadata: Dict[str, Any]
    ) -> Optional[app_schema.AppDefinition]:
        """See [DB.update_app_metadata][trulens.core.database.base.DB.update_app_metadata]."""

        def nested_update(metadata: dict, update: dict):
            for k, v in update.items():
                if isinstance(v, dict) and k in metadata:
                    nested_update(metadata[k], v)
                else:
                    metadata[k] = v

        with self.session.begin() as session:
            if (
                _app := session.query(self.orm.AppDefinition)
                .filter_by(app_id=app_id)
                .first()
            ):
                app_json = json.loads(_app.app_json)
                if "metadata" not in app_json:
                    app_json["metadata"] = {}
                nested_update(app_json["metadata"], metadata)
                _app.app_json = json.dumps(app_json)

    def get_apps(
        self, app_name: Optional[types_schema.AppName] = None
    ) -> Iterable[serial_utils.JSON]:
        """See [DB.get_apps][trulens.core.database.base.DB.get_apps]."""

        with self.session.begin() as session:
            q = sa.select(self.orm.AppDefinition)
            if app_name:
                q = q.filter_by(app_name=app_name)
            app_defs = (row[0] for row in session.execute(q))
            for _app in app_defs:
                yield json.loads(_app.app_json)

    def insert_app(self, app: app_schema.AppDefinition) -> types_schema.AppID:
        """See [DB.insert_app][trulens.core.database.base.DB.insert_app]."""

        # TODO: thread safety

        with self.session.begin() as session:
            if (
                _app := session.query(self.orm.AppDefinition)
                .filter_by(app_id=app.app_id)
                .first()
            ):
                _app.app_json = app.model_dump_json()
            else:
                _app = self.orm.AppDefinition.parse(
                    app, redact_keys=self.redact_keys
                )
                session.merge(_app)  # .add was not thread safe

            logger.info(
                "%s added app %s", text_utils.UNICODE_CHECK, _app.app_id
            )

            return _app.app_id

    def delete_app(self, app_id: types_schema.AppID) -> None:
        """
        Deletes an app from the database based on its app_id.

        Args:
            app_id (schema.AppID): The unique identifier of the app to be deleted.
        """
        with self.session.begin() as session:
            _app = (
                session.query(self.orm.AppDefinition)
                .filter_by(app_id=app_id)
                .first()
            )
            if _app:
                session.delete(_app)
                logger.info(f"{text_utils.UNICODE_CHECK} deleted app {app_id}")
            else:
                logger.warning(f"App {app_id} not found for deletion.")

    def insert_feedback_definition(
        self, feedback_definition: feedback_schema.FeedbackDefinition
    ) -> types_schema.FeedbackDefinitionID:
        """See [DB.insert_feedback_definition][trulens.core.database.base.DB.insert_feedback_definition]."""

        # TODO: thread safety

        with self.session.begin() as session:
            if (
                _fb_def := session.query(self.orm.FeedbackDefinition)
                .filter_by(
                    feedback_definition_id=feedback_definition.feedback_definition_id
                )
                .first()
            ):
                _fb_def.app_json = feedback_definition.model_dump_json()
            else:
                _fb_def = self.orm.FeedbackDefinition.parse(
                    feedback_definition, redact_keys=self.redact_keys
                )
                session.merge(_fb_def)  # .add was not thread safe

            logger.info(
                "%s added feedback definition %s",
                text_utils.UNICODE_CHECK,
                _fb_def.feedback_definition_id,
            )

            return _fb_def.feedback_definition_id

    def get_feedback_defs(
        self,
        feedback_definition_id: Optional[
            types_schema.FeedbackDefinitionID
        ] = None,
    ) -> pd.DataFrame:
        """See [DB.get_feedback_defs][trulens.core.database.base.DB.get_feedback_defs]."""

        with self.session.begin() as session:
            q = sa.select(self.orm.FeedbackDefinition)
            if feedback_definition_id:
                q = q.filter_by(feedback_definition_id=feedback_definition_id)
            fb_defs = (row[0] for row in session.execute(q))
            return pd.DataFrame(
                data=(
                    (fb.feedback_definition_id, json.loads(fb.feedback_json))
                    for fb in fb_defs
                ),
                columns=["feedback_definition_id", "feedback_json"],
            )

    def insert_feedback(
        self, feedback_result: feedback_schema.FeedbackResult
    ) -> types_schema.FeedbackResultID:
        """See [DB.insert_feedback][trulens.core.database.base.DB.insert_feedback]."""
        if is_otel_tracing_enabled():
            raise RuntimeError("Not supported with OTel tracing enabled!")

        # TODO: thread safety

        _feedback_result = self.orm.FeedbackResult.parse(
            feedback_result, redact_keys=self.redact_keys
        )
        with self.session.begin() as session:
            # The Snowflake stored procedure connector isn't currently capable
            # of handling None qmark-bound to an `INSERT INTO` or `UPDATE`
            # statement for nullable numeric columns. Thus, as a hack, we get
            # around this by first inserting a non-null value then updating it
            # to a null value.
            use_snowflake_hack = (
                self.engine.dialect.name == "snowflake"
                and _feedback_result.result is None
            )
            if not use_snowflake_hack:
                session.merge(_feedback_result)
            else:
                _feedback_result.result = -1
                session.merge(_feedback_result)
                _feedback_result.result = None
                session.flush()  # Ensure the merge is executed before the update.
                session.execute(
                    sql_text(
                        """
                    UPDATE trulens_feedbacks
                    SET result=NULL
                    WHERE trulens_feedbacks.feedback_result_id = :feedback_result_id
                        """.replace("\n", " ")
                    ),
                    {"feedback_result_id": feedback_result.feedback_result_id},
                )
                session.flush()

            status = feedback_schema.FeedbackResultStatus(
                _feedback_result.status
            )

            if status == feedback_schema.FeedbackResultStatus.DONE:
                icon = text_utils.UNICODE_CHECK
            elif status == feedback_schema.FeedbackResultStatus.RUNNING:
                icon = text_utils.UNICODE_HOURGLASS
            elif status == feedback_schema.FeedbackResultStatus.NONE:
                icon = text_utils.UNICODE_CLOCK
            elif status == feedback_schema.FeedbackResultStatus.FAILED:
                icon = text_utils.UNICODE_STOP
            else:
                icon = "???"

            logger.info(
                "%s feedback result %s %s %s",
                icon,
                _feedback_result.name,
                status.name,
                _feedback_result.feedback_result_id,
            )

            return _feedback_result.feedback_result_id

    def batch_insert_feedback(
        self, feedback_results: List[feedback_schema.FeedbackResult]
    ) -> List[types_schema.FeedbackResultID]:
        """See [DB.batch_insert_feedback][trulens.core.database.base.DB.batch_insert_feedback]."""
        if is_otel_tracing_enabled():
            raise RuntimeError("Not supported with OTel tracing enabled!")
        # The Snowflake stored procedure connector isn't currently capable of
        # handling None qmark-bound to an `INSERT INTO` or `UPDATE` statement
        # for nullable numeric columns. Thus, as a hack, we get around this by
        # first inserting a non-null value then updating it to a null value.
        if self.engine.dialect == "snowflake" and any([
            curr.result is None for curr in feedback_results
        ]):
            ret = []
            for curr in feedback_results:
                ret.append(self.insert_feedback(curr))
            return ret
        with self.session.begin() as session:
            feedback_results_list = [
                self.orm.FeedbackResult.parse(f, redact_keys=self.redact_keys)
                for f in feedback_results
            ]
            session.add_all(feedback_results_list)
            return [f.feedback_result_id for f in feedback_results_list]

    def _feedback_query(
        self,
        count_by_status: bool = False,
        shuffle: bool = False,
        record_id: Optional[types_schema.RecordID] = None,
        feedback_result_id: Optional[types_schema.FeedbackResultID] = None,
        feedback_definition_id: Optional[
            types_schema.FeedbackDefinitionID
        ] = None,
        status: Optional[
            Union[
                feedback_schema.FeedbackResultStatus,
                Sequence[feedback_schema.FeedbackResultStatus],
            ]
        ] = None,
        last_ts_before: Optional[datetime] = None,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
        run_location: Optional[feedback_schema.FeedbackRunLocation] = None,
    ):
        if count_by_status:
            q = sa.select(
                self.orm.FeedbackResult.status,
                sa.func.count(self.orm.FeedbackResult.feedback_result_id),
            ).group_by(self.orm.FeedbackResult.status)
        else:
            q = sa.select(self.orm.FeedbackResult)

        if record_id:
            q = q.filter_by(record_id=record_id)

        if feedback_result_id:
            q = q.filter_by(feedback_result_id=feedback_result_id)

        if feedback_definition_id:
            q = q.filter_by(feedback_definition_id=feedback_definition_id)

        if (
            run_location is None
            or run_location == feedback_schema.FeedbackRunLocation.IN_APP
        ):
            # For legacy reasons, we handle the IN_APP and NULL/None case as the same.
            q = q.filter(
                sa.or_(
                    self.orm.FeedbackDefinition.run_location.is_(None),
                    self.orm.FeedbackDefinition.run_location
                    == feedback_schema.FeedbackRunLocation.IN_APP.value,
                )
            )
        else:
            q = q.filter(
                self.orm.FeedbackDefinition.run_location == run_location.value
            )
        q = q.filter(
            self.orm.FeedbackResult.feedback_definition_id
            == self.orm.FeedbackDefinition.feedback_definition_id
        )

        if status:
            if isinstance(status, feedback_schema.FeedbackResultStatus):
                status = [status.value]
            q = q.filter(
                self.orm.FeedbackResult.status.in_([s.value for s in status])
            )
        if last_ts_before:
            q = q.filter(
                self.orm.FeedbackResult.last_ts < last_ts_before.timestamp()
            )

        if offset is not None:
            q = q.offset(offset)

        if limit is not None:
            q = q.limit(limit)

        if shuffle:
            q = q.order_by(sa.func.random())

        return q

    def get_feedback_count_by_status(
        self,
        record_id: Optional[types_schema.RecordID] = None,
        feedback_result_id: Optional[types_schema.FeedbackResultID] = None,
        feedback_definition_id: Optional[
            types_schema.FeedbackDefinitionID
        ] = None,
        status: Optional[
            Union[
                feedback_schema.FeedbackResultStatus,
                Sequence[feedback_schema.FeedbackResultStatus],
            ]
        ] = None,
        last_ts_before: Optional[datetime] = None,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
        shuffle: bool = False,
        run_location: Optional[feedback_schema.FeedbackRunLocation] = None,
    ) -> Dict[feedback_schema.FeedbackResultStatus, int]:
        """See [DB.get_feedback_count_by_status][trulens.core.database.base.DB.get_feedback_count_by_status]."""
        if is_otel_tracing_enabled():
            raise RuntimeError("Not supported with OTel tracing enabled!")
        with self.session.begin() as session:
            q = self._feedback_query(
                count_by_status=True,
                **python_utils.locals_except("self", "session"),
            )
            results = session.execute(q)

            return {
                feedback_schema.FeedbackResultStatus(row[0]): row[1]
                for row in results
            }

    def get_feedback(
        self,
        record_id: Optional[types_schema.RecordID] = None,
        feedback_result_id: Optional[types_schema.FeedbackResultID] = None,
        feedback_definition_id: Optional[
            types_schema.FeedbackDefinitionID
        ] = None,
        status: Optional[
            Union[
                feedback_schema.FeedbackResultStatus,
                Sequence[feedback_schema.FeedbackResultStatus],
            ]
        ] = None,
        last_ts_before: Optional[datetime] = None,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
        shuffle: Optional[bool] = False,
        run_location: Optional[feedback_schema.FeedbackRunLocation] = None,
    ) -> pd.DataFrame:
        """See [DB.get_feedback][trulens.core.database.base.DB.get_feedback]."""
        if is_otel_tracing_enabled():
            raise RuntimeError("Not supported with OTel tracing enabled!")
        with self.session.begin() as session:
            q = self._feedback_query(
                **python_utils.locals_except("self", "session")
            )

            results = (row[0] for row in session.execute(q))

            return _extract_feedback_results(results)

    def _json_extract_otel(self, column: str, path: str) -> sa.Column:
        """Helper function to extract JSON values from a JSON column in the Event table.

        Args:
            column: The name of the JSON column to extract from (e.g. 'record_attributes', 'record', etc.)
            path: The JSON path to extract from the column

        Returns:
            A SQLAlchemy column expression that extracts the value at the given path

        Raises:
            ValueError: If the column doesn't exist or is not a JSON column
        """
        if not hasattr(self.orm.Event, column):
            raise ValueError(f"Column {column} not found in Event table")

        column_obj = getattr(self.orm.Event, column)
        if not isinstance(column_obj.type, sa.JSON):
            raise ValueError(f"Column {column} is not a JSON column")

        dialect = self.engine.dialect.name
        if dialect == "snowflake":
            return sa.func.json_extract_path_text(column_obj, f'"{path}"')
        elif dialect == "postgresql":
            return sa.func.json_extract_path_text(column_obj, path)
        else:
            # SQLite and MySQL use json_extract with JSONPath syntax
            return sa.func.json_extract(column_obj, f'$."{path}"')

    def _get_paginated_record_ids_otel(
        self,
        app_ids: Optional[List[str]] = None,
        app_name: Optional[types_schema.AppName] = None,
        app_version: Optional[types_schema.AppVersion] = None,
        app_versions: Optional[List[types_schema.AppVersion]] = None,
        run_name: Optional[types_schema.RunName] = None,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> sa.Select:
        """Get a paginated query for record IDs from the OTEL event table.

        Args:
            app_ids: List of app IDs to filter by. Defaults to None.
            app_name: App name to filter by. Defaults to None.
            app_version: App version to filter by. Defaults to None.
            app_versions: List of app versions to filter by. Defaults to None.
            run_name: Run name to filter by. Defaults to None.
            offset: Offset for pagination. Defaults to None.
            limit: Limit for pagination. Defaults to None.

        Returns:
            A SQLAlchemy select statement for record IDs with pagination applied.
        """

        # Base select statement for record IDs and timestamps
        stmt = sa.select(
            self._json_extract_otel(
                "record_attributes", SpanAttributes.RECORD_ID
            ).label("record_id"),
            sa.func.min(self.orm.Event.start_timestamp).label(
                "min_start_timestamp"
            ),
        )

        # Common conditions for filtering
        conditions = [
            # Filter for record_root span type events
            self._json_extract_otel(
                "record_attributes", SpanAttributes.SPAN_TYPE
            )
            == SpanAttributes.SpanType.RECORD_ROOT.value,
            # Filter out NULL and empty record IDs
            self._json_extract_otel(
                "record_attributes", SpanAttributes.RECORD_ID
            ).isnot(None),
            self._json_extract_otel(
                "record_attributes", SpanAttributes.RECORD_ID
            )
            != "",
        ]

        # Add app_name filter if provided
        if app_name:
            app_name_expr = self._json_extract_otel(
                "resource_attributes", ResourceAttributes.APP_NAME
            )
            conditions.append(app_name_expr == app_name)

        if app_version:
            app_version_expr = self._json_extract_otel(
                "resource_attributes", ResourceAttributes.APP_VERSION
            )
            conditions.append(app_version_expr == app_version)

        if app_versions:
            app_version_expr = self._json_extract_otel(
                "resource_attributes", ResourceAttributes.APP_VERSION
            )
            conditions.append(app_version_expr.in_(app_versions))

        if app_ids:
            app_id_expr = self._json_extract_otel(
                "resource_attributes", ResourceAttributes.APP_ID
            )
            conditions.append(
                sa.or_(
                    app_id_expr.in_(app_ids),
                    app_id_expr.is_(None),
                    app_id_expr == "",
                )
            )

        if run_name:
            run_name_expr = self._json_extract_otel(
                "record_attributes", SpanAttributes.RUN_NAME
            )
            conditions.append(run_name_expr == run_name)

        # Apply all conditions
        stmt = stmt.where(sa.and_(*conditions))

        # Group by record_id to get unique records
        stmt = stmt.group_by("record_id")

        # Order by timestamp ascending (oldest first)
        stmt = stmt.order_by(sa.asc("min_start_timestamp"))

        # Apply pagination
        if limit is not None:
            stmt = stmt.limit(limit)
        if offset is not None:
            stmt = stmt.offset(offset)

        return stmt

    def _get_records_and_feedback_otel(
        self,
        app_ids: Optional[List[str]] = None,
        app_name: Optional[types_schema.AppName] = None,
        app_version: Optional[types_schema.AppVersion] = None,
        app_versions: Optional[List[types_schema.AppVersion]] = None,
        run_name: Optional[types_schema.RunName] = None,
        record_ids: Optional[List[types_schema.RecordID]] = None,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> Tuple[pd.DataFrame, Sequence[str]]:
        """Get records and feedback from the OTEL event table.

        This method builds a records dataframe from the EVENT table that mirrors
        the structure of the pre-OTEL ORM's get_records_and_feedback method.

        Args:
            app_ids: List of app IDs to filter by. Defaults to None.
            app_name: App name to filter by. Defaults to None.
            app_version: App version to filter by. Defaults to None.
            app_versions: List of app versions to filter by. Defaults to None.
            run_name: Run name to filter by. Defaults to None.
            record_ids: List of record IDs to filter by. Defaults to None.
            offset: Offset for pagination. Defaults to None.
            limit: Limit for pagination. Defaults to None.

        Returns:
            A tuple of (records dataframe, feedback column names).
        """
        with self.session.begin() as session:
            # Get paginated record IDs
            if record_ids is None:
                record_id_subquery = self._get_paginated_record_ids_otel(
                    app_ids=app_ids,
                    app_name=app_name,
                    app_version=app_version,
                    app_versions=app_versions,
                    run_name=run_name,
                    offset=offset,
                    limit=limit,
                )
                record_ids_sql = sa.select(record_id_subquery.c.record_id)
            else:
                record_ids_sql = record_ids

            # Now get all events for those record IDs
            stmt = sa.select(self.orm.Event).where(
                self._json_extract_otel(
                    "record_attributes", SpanAttributes.RECORD_ID
                ).in_(record_ids_sql)
            )

            # Execute query
            events = session.execute(stmt).scalars().all()
            return self._get_records_and_feedback_otel_from_events(
                events=events,
                app_ids=app_ids,
                app_name=app_name,
            )

    def get_records_and_feedback(
        self,
        app_ids: Optional[List[types_schema.AppID]] = None,
        app_name: Optional[types_schema.AppName] = None,
        app_version: Optional[types_schema.AppVersion] = None,
        app_versions: Optional[List[types_schema.AppVersion]] = None,
        run_name: Optional[types_schema.RunName] = None,
        record_ids: Optional[List[types_schema.RecordID]] = None,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> Tuple[pd.DataFrame, Sequence[str]]:
        """See [DB.get_records_and_feedback][trulens.core.database.base.DB.get_records_and_feedback]."""
        if app_ids:
            logger.warning(
                "`app_ids` is deprecated. Please use `app_name`, and `app_versions` instead."
            )
        if app_version:
            logger.warning(
                "`app_version` is deprecated. Please use `app_versions` instead."
            )
        if is_otel_tracing_enabled():
            return self._get_records_and_feedback_otel(
                app_ids=app_ids,
                app_name=app_name,
                app_version=app_version,
                app_versions=app_versions,
                run_name=run_name,
                record_ids=record_ids,
                offset=offset,
                limit=limit,
            )
        elif run_name:
            raise RuntimeError(
                "`run_name` is not supported in the pre-OTel implementation."
            )

        # Original implementation for pre-OTEL ORM

        # TODO: Add pagination to this method. Currently the joinedload in
        # select below disables lazy loading of records which will be a problem
        # for large databases without the use of pagination.

        if record_ids is not None:
            raise NotImplementedError(
                "`record_ids` is not supported in the pre-OTEL implementation."
            )

        with self.session.begin() as session:
            stmt = sa.select(self.orm.Record)
            # NOTE: We are selecting records here because offset and limit need
            # to be with respect to those rows instead of AppDefinition or
            # FeedbackResult rows.

            if app_ids:
                stmt = stmt.where(self.orm.Record.app_id.in_(app_ids))

            if app_name:
                # stmt = stmt.options(joinedload(self.orm.Record.app))
                stmt = stmt.join(self.orm.Record.app).filter(
                    self.orm.AppDefinition.app_name == app_name
                )

            if app_version:
                # stmt = stmt.options(joinedload(self.orm.Record.app))
                stmt = stmt.join(self.orm.Record.app).filter(
                    self.orm.AppDefinition.app_version == app_version
                )

            if app_versions:
                stmt = stmt.join(self.orm.Record.app).filter(
                    self.orm.AppDefinition.app_version.in_(app_versions)
                )

            stmt = stmt.options(joinedload(self.orm.Record.feedback_results))
            stmt = stmt.options(joinedload(self.orm.Record.app))
            # NOTE(piotrm): The joinedload here makes it so that the
            # feedback_results and app definitions get loaded eagerly instead if lazily when
            # accessed later.

            # TODO(piotrm): The subsequent logic in helper methods end up
            # reading all of the records and feedback_results in order to create
            # a DataFrame so there is no reason to not eagerly get all of this
            # data. Ideally, though, we would be making some sort of lazy
            # DataFrame and then could use the lazy join feature of sqlalchemy.

            stmt = stmt.order_by(
                self.orm.Record.ts.desc(), self.orm.Record.record_id
            )
            # NOTE: feedback_results order is governed by the order_by on the
            # orm.FeedbackResult.record backref definition. Here, we need to
            # order Records as we did not use an auto join to retrieve them. If
            # records were to be retrieved from AppDefinition.records via auto
            # join, though, the orm backref ordering would be able to take hold.

            # Apply pagination
            if limit is not None:
                stmt = stmt.limit(limit)
            if offset is not None:
                stmt = stmt.offset(offset)

            ex = session.execute(stmt).unique()
            # unique needed for joinedload above.

            records = [rec[0] for rec in ex]
            # TODO: Make the iteration of records lazy in some way. See
            # TODO(piotrm) above.

            return AppsExtractor().get_df_and_cols(records=records)

    def insert_ground_truth(
        self, ground_truth: groundtruth_schema.GroundTruth
    ) -> types_schema.GroundTruthID:
        """See [DB.insert_ground_truth][trulens.core.database.base.DB.insert_ground_truth]."""

        # TODO: thread safety
        with self.session.begin() as session:
            if (
                _ground_truth := session.query(self.orm.GroundTruth)
                .filter_by(ground_truth_id=ground_truth.ground_truth_id)
                .first()
            ):
                # Update the existing record for idempotency
                _ground_truth.ground_truth_json = ground_truth.model_dump_json()
            else:
                _ground_truth = self.orm.GroundTruth.parse(
                    ground_truth, redact_keys=self.redact_keys
                )

                session.merge(_ground_truth)

            logger.info(
                f"{text_utils.UNICODE_CHECK} added ground truth {_ground_truth.ground_truth_id}"
            )

            return _ground_truth.ground_truth_id

    def batch_insert_ground_truth(
        self, ground_truths: List[groundtruth_schema.GroundTruth]
    ) -> List[types_schema.GroundTruthID]:
        """See [DB.batch_insert_ground_truth][trulens.core.database.base.DB.batch_insert_ground_truth]."""
        with self.session.begin() as session:
            ground_truth_ids = [gt.ground_truth_id for gt in ground_truths]

            # Fetch existing GroundTruth records that match these ids in one query
            existing_ground_truths = (
                session.query(self.orm.GroundTruth)
                .filter(
                    self.orm.GroundTruth.ground_truth_id.in_(ground_truth_ids)
                )
                .all()
            )

            existing_ground_truth_dict = {
                gt.ground_truth_id: gt for gt in existing_ground_truths
            }

            ground_truths_to_insert = []
            for ground_truth in ground_truths:
                if ground_truth.ground_truth_id in existing_ground_truth_dict:
                    existing_record = existing_ground_truth_dict[
                        ground_truth.ground_truth_id
                    ]
                    # Update the existing record for idempotency
                    existing_record.ground_truth_json = (
                        ground_truth.model_dump_json()
                    )
                else:
                    new_ground_truth = self.orm.GroundTruth.parse(
                        ground_truth, redact_keys=self.redact_keys
                    )
                    ground_truths_to_insert.append(new_ground_truth)

            session.add_all(ground_truths_to_insert)
            return [gt.ground_truth_id for gt in ground_truths]

    def get_ground_truth(
        self, ground_truth_id: str | None = None
    ) -> Optional[serial_utils.JSONized]:
        """See [DB.get_ground_truth][trulens.core.database.base.DB.get_ground_truth]."""

        with self.session.begin() as session:
            if (
                _ground_truth := session.query(self.orm.GroundTruth)
                .filter_by(ground_truth_id=ground_truth_id)
                .first()
            ):
                return json.loads(_ground_truth)

    def get_ground_truths_by_dataset(
        self, dataset_name: str
    ) -> pd.DataFrame | None:
        """See [DB.get_ground_truths_by_dataset][trulens.core.database.base.DB.get_ground_truths_by_dataset]."""
        with self.session.begin() as session:
            q = sa.select(self.orm.Dataset)
            all_datasets = (row[0] for row in session.execute(q))
            df = None
            for dataset in all_datasets:
                dataset_json = json.loads(dataset.dataset_json)
                if (
                    "name" in dataset_json
                    and dataset_json["name"] == dataset_name
                ):
                    q = sa.select(self.orm.GroundTruth).filter_by(
                        dataset_id=dataset.dataset_id
                    )
                    results = (row[0] for row in session.execute(q))

                    if df is None:
                        df = _extract_ground_truths(results)
                    else:
                        df = pd.concat([df, _extract_ground_truths(results)])
            return df
            # TODO: use a generator instead of a list? (for large datasets)

    def get_virtual_ground_truth(
        self,
        user_table_name: str,
        user_schema_mapping: Dict[str, str],
        user_schema_name: Optional[str] = None,
    ) -> pd.DataFrame:
        virtual_gt_schema_mapping: groundtruth_schema.VirtualGroundTruthSchemaMapping = groundtruth_schema.VirtualGroundTruthSchemaMapping.validate_mapping(
            user_schema_mapping
        )
        with self.session.begin() as session:
            try:
                # Dynamically construct the column mappings
                inspector = inspect(session.get_bind())
                user_columns = inspector.get_columns(
                    user_table_name, schema=user_schema_name
                )

                # Dynamically define the table with its columns
                user_table = sa.Table(
                    user_table_name,
                    sa.MetaData(),
                    *[
                        sa.Column(
                            col["name"], col["type"], nullable=col["nullable"]
                        )
                        for col in user_columns
                    ],
                    schema=user_schema_name,
                )

                # Ensure all required columns exist
                required_columns = {
                    getattr(virtual_gt_schema_mapping, field)
                    for field in virtual_gt_schema_mapping.model_dump()
                    if getattr(virtual_gt_schema_mapping, field)
                }

                existing_columns = {col.name for col in user_table.columns}
                missing_columns = required_columns - existing_columns
                if missing_columns:
                    raise ValueError(
                        f"Missing required columns in user table '{user_table_name}': {', '.join(missing_columns)}"
                    )

                # Create a query to fetch the mapped data
                column_mapping = {
                    field: user_table.c[
                        getattr(virtual_gt_schema_mapping, field)
                    ]
                    for field in virtual_gt_schema_mapping.model_dump()
                    if getattr(virtual_gt_schema_mapping, field)
                }

                query = sa.select(*column_mapping.values())
                results = session.execute(query).fetchall()

                # Convert results to DataFrame
                if not results:
                    return pd.DataFrame()
                return pd.DataFrame(
                    data=results, columns=list(column_mapping.keys())
                )
            except Exception as e:
                raise ValueError(
                    f"Failed to fetch virtual ground truth: {str(e)}"
                )

    def insert_dataset(
        self, dataset: dataset_schema.Dataset
    ) -> types_schema.DatasetID:
        """See [DB.insert_dataset][trulens.core.database.base.DB.insert_dataset]."""

        with self.session.begin() as session:
            if (
                _dataset := session.query(self.orm.Dataset)
                .filter_by(dataset_id=dataset.dataset_id)
                .first()
            ):
                # Update the existing record for idempotency
                _dataset.dataset_json = dataset.model_dump_json()
            else:
                _dataset = self.orm.Dataset.parse(
                    dataset, redact_keys=self.redact_keys
                )
                session.merge(_dataset)

            logger.info(
                f"{text_utils.UNICODE_CHECK} added dataset {_dataset.dataset_id}"
            )

            return _dataset.dataset_id

    def get_datasets(self) -> pd.DataFrame:
        """See [DB.get_datasets][trulens.core.database.base.DB.get_datasets]."""

        with self.session.begin() as session:
            results = session.query(self.orm.Dataset)

            return pd.DataFrame(
                data=((ds.dataset_id, ds.name, ds.meta) for ds in results),
                columns=["dataset_id", "name", "meta"],
            )

    def insert_event(self, event: Event) -> types_schema.EventID:
        """See [DB.insert_event][trulens.core.database.base.DB.insert_event]."""
        return self.insert_events([event])[0]

    def insert_events(self, events: List[Event]) -> List[types_schema.EventID]:
        """See [DB.insert_events][trulens.core.database.base.DB.insert_events]."""
        with self.session.begin() as session:
            events_to_insert = [
                self.orm.Event.parse(event, redact_keys=self.redact_keys)
                for event in events
            ]
            session.add_all(events_to_insert)
            ret = [event.event_id for event in events_to_insert]
            for curr in ret:
                logger.info(f"{text_utils.UNICODE_CHECK} added event {curr}")
            return ret

    def get_events(
        self,
        app_name: Optional[str],
        app_version: Optional[str],
        record_ids: Optional[List[str]],
        start_time: Optional[datetime],
    ) -> pd.DataFrame:
        """See [DB.get_events][trulens.core.database.base.DB.get_events]."""
        with self.session.begin() as session:
            where_clauses = []
            if app_name is not None:
                app_name_expr = self._json_extract_otel(
                    "resource_attributes", ResourceAttributes.APP_NAME
                )
                where_clauses.append(app_name_expr == app_name)
            if app_version is not None:
                app_version_expr = self._json_extract_otel(
                    "resource_attributes", ResourceAttributes.APP_VERSION
                )
                where_clauses.append(app_version_expr == app_version)
            if record_ids is not None:
                record_id_expr = self._json_extract_otel(
                    "record_attributes", SpanAttributes.RECORD_ID
                )
                where_clauses.append(record_id_expr.in_(record_ids))
            if start_time is not None:
                where_clauses.append(self.orm.Event.timestamp >= start_time)

            if len(where_clauses) == 0:
                q = sa.select(self.orm.Event)
            elif len(where_clauses) == 1:
                q = sa.select(self.orm.Event).where(where_clauses[0])
            else:
                q = sa.select(self.orm.Event).where(sa.and_(*where_clauses))
            return pd.read_sql(q, session.bind)

    def _extract_namespaced_attributes(
        self, record_attributes: Dict[str, Any], namespace_prefix: str
    ) -> Dict[str, Any]:
        """Extract attributes that are namespaced under a given prefix.

        Args:
            record_attributes: Dictionary of all span attributes
            namespace_prefix: The namespace prefix to look for (e.g., "ai.observability.call.kwargs")

        Returns:
            Dictionary with namespace-stripped keys and their values
        """
        prefix_with_dot = namespace_prefix + "."
        return {
            key[len(prefix_with_dot) :]: value
            for key, value in record_attributes.items()
            if key.startswith(prefix_with_dot)
        }


# Use this Perf for missing Perfs.
# TODO: Migrate the database instead.
def _make_no_perf():
    # Def to avoid circular imports.
    return base_schema.Perf.min().model_dump()


def _extract_feedback_results(
    results: Iterable["db_orm.FeedbackResult"],
) -> pd.DataFrame:
    def _extract(_result: "db_orm.FeedbackResult"):
        app_json = json.loads(_result.record.app.app_json)
        _type = app_schema.AppDefinition.model_validate(app_json).root_class

        return (
            _result.record_id,
            _result.feedback_result_id,
            _result.feedback_definition_id,
            _result.last_ts,
            feedback_schema.FeedbackResultStatus(_result.status),
            _result.error,
            _result.name,
            _result.result,
            _result.multi_result,
            _result.cost_json,  # why is cost_json not parsed?
            json.loads(_result.record.perf_json)
            if _result.record.perf_json
            != legacy_migration.MIGRATION_UNKNOWN_STR
            else _make_no_perf(),
            json.loads(_result.calls_json)["calls"],
            json.loads(_result.feedback_definition.feedback_json)
            if _result.feedback_definition is not None
            else None,
            json.loads(_result.record.record_json),
            app_json,
            _type,
        )

    df = pd.DataFrame(
        data=(_extract(r) for r in results),
        columns=[
            "record_id",
            "feedback_result_id",
            "feedback_definition_id",
            "last_ts",
            "status",
            "error",
            "fname",
            "result",
            "multi_result",
            "cost_json",
            "perf_json",
            "calls_json",
            "feedback_json",
            "record_json",
            "app_json",
            "type",
        ],
    )
    df["latency"] = _extract_latency(df["perf_json"])
    df = pd.concat([df, _extract_tokens_and_cost(df["cost_json"])], axis=1)
    return df


def _extract_latency(
    series: pd.Series,
) -> pd.Series:
    def _extract(perf_json: Union[str, dict, base_schema.Perf]) -> float:
        if perf_json == legacy_migration.MIGRATION_UNKNOWN_STR:
            return np.nan

        if isinstance(perf_json, str):
            perf_json = json.loads(perf_json)

        if isinstance(perf_json, dict):
            perf_json = base_schema.Perf.model_validate(perf_json)

        if isinstance(perf_json, base_schema.Perf):
            return (
                perf_json.latency.seconds + perf_json.latency.microseconds / 1e6
            )

        if perf_json is None:
            return 0

        raise ValueError(f"Failed to parse perf_json: {perf_json}")

    return series.apply(_extract)


def _extract_tokens_and_cost(cost_json: pd.Series) -> pd.DataFrame:
    def _extract(_cost_json: Union[str, dict]) -> Tuple[int, float, str]:
        if isinstance(_cost_json, str):
            _cost_json = json.loads(_cost_json)
        if _cost_json is not None:
            cost = base_schema.Cost(**_cost_json)
        else:
            cost = base_schema.Cost()
        return cost.n_tokens, cost.cost, cost.cost_currency

    return pd.DataFrame(
        data=(_extract(c) for c in cost_json),
        columns=["total_tokens", "total_cost", "cost_currency"],
    )


def _extract_ground_truths(
    results: Iterable["db_orm.GroundTruth"],
) -> pd.DataFrame:
    def _extract(_result: "db_orm.GroundTruth"):
        ground_truth_json = json.loads(_result.ground_truth_json)

        return (
            _result.ground_truth_id,
            _result.dataset_id,
            ground_truth_json["query"],
            ground_truth_json["query_id"],
            ground_truth_json["expected_response"],
            ground_truth_json["expected_chunks"],
            ground_truth_json["meta"],
        )

    return pd.DataFrame(
        data=(_extract(r) for r in results),
        columns=[
            "ground_truth_id",
            "dataset_id",
            "query",
            "query_id",
            "expected_response",
            "expected_chunks",
            "meta",
        ],
    )


class AppsExtractor(core_db.BaseAppsExtractor):
    """Utilities for creating dataframes from orm instances."""

    def __init__(self):
        self.feedback_columns = set()

    def get_df_and_cols(
        self,
        apps: Optional[List["db_orm.ORM.AppDefinition"]] = None,
        records: Optional[List["db_orm.ORM.Record"]] = None,
    ) -> Tuple[pd.DataFrame, Sequence[str]]:
        """Produces a records dataframe which joins in information from apps and
        feedback results.

        Args:
            apps: If given, includes all records of all of the apps in this
                iterable.

            records: If given, includes only these records. Mutually exclusive
                with `apps`.
        """

        assert (
            apps is None or records is None
        ), "`apps` and `records` are mutually exclusive"

        if apps is not None:
            df = pd.concat(self.extract_apps(apps))

        elif records is not None:
            apps = {record.app for record in records}
            df = pd.concat(self.extract_apps(apps=apps, records=records))

        else:
            raise ValueError("'apps` or `records` must be provided")

        df["latency"] = _extract_latency(df["perf_json"])
        df.reset_index(
            drop=True, inplace=True
        )  # prevent index mismatch on the horizontal concat that follows
        df = pd.concat([df, _extract_tokens_and_cost(df["cost_json"])], axis=1)
        df["app_json"] = df["app_json"].apply(json.loads)
        df["record_json"] = df["record_json"].apply(json.loads)
        df["input"] = df["input"].apply(json.loads)
        df["output"] = df["output"].apply(json.loads)

        return df, list(self.feedback_columns)

    def extract_apps(
        self,
        apps: Iterable["db_orm.ORM.AppDefinition"],
        records: Optional[List["db_orm.ORM.Record"]] = None,
    ) -> Iterable[pd.DataFrame]:
        """
        Creates record rows with app information.

        TODO: The means for enumerating records in this method is not ideal as
        it does a lot of filtering.
        """

        yield pd.DataFrame(
            [], columns=self.app_cols + self.rec_cols
        )  # prevent empty iterator
        for _app in apps:
            try:
                if records is None:
                    # If records not provided, get all of them for `_app`.
                    _recs = _app.records
                else:
                    # Otherwise get only the ones in `records`. WARNING: Avoid
                    # using _app.records here as doing so might get all of the
                    # records even the ones not in `records`
                    _recs = (
                        record
                        for record in records
                        if record.app_id == _app.app_id
                    )

                if _recs:
                    df = pd.DataFrame(data=self.extract_records(_recs))

                    for col in self.app_cols:
                        if col == "type":
                            # Previous DBs did not contain entire app so we cannot
                            # deserialize AppDefinition here unless we fix prior DBs
                            # in migration. Because of this, loading just the
                            # `root_class` here.

                            df[col] = str(
                                pyschema_utils.Class.model_validate(
                                    json.loads(_app.app_json).get("root_class")
                                )
                            )

                        else:
                            df[col] = getattr(_app, col)

                    df["app_name"] = _app.app_name
                    df["app_version"] = _app.app_version
                    yield df
            except OperationalError as e:
                print(
                    "Error encountered while attempting to retrieve an app. "
                    "This issue may stem from a corrupted database."
                )
                print(f"Error details: {e}")

    def extract_records(
        self, records: Iterable["db_orm.ORM.Record"]
    ) -> Iterable[pd.Series]:
        if is_otel_tracing_enabled():
            raise RuntimeError("Not supported with OTel tracing enabled!")
        for _rec in records:
            calls = defaultdict(list)
            values = defaultdict(list)
            feedback_cost = {}
            try:
                for _res in _rec.feedback_results:
                    calls[_res.name].append(
                        json.loads(_res.calls_json)["calls"]
                    )

                    feedback_usage = json.loads(_res.cost_json)
                    cost_currency = feedback_usage.get("cost_currency", "USD")

                    if "cost" in feedback_usage:
                        feedback_cost[
                            f"{_res.name} feedback cost in {cost_currency}"
                        ] = feedback_usage["cost"]

                    if (
                        _res.multi_result not in [None, "null", "None"]
                        and (multi_result := json.loads(_res.multi_result))
                        is not None
                    ):
                        for key, val in multi_result.items():
                            if (
                                val is not None
                            ):  # avoid getting Nones into np.mean
                                name = f"{_res.name}{core_db.MULTI_CALL_NAME_DELIMITER}{key}"
                                values[name] = val
                                self.feedback_columns.add(name)
                    elif (
                        _res.result is not None
                    ):  # avoid getting Nones into np.mean
                        values[_res.name].append(_res.result)
                        self.feedback_columns.add(_res.name)

                row = {
                    **{k: np.mean(v) for k, v in values.items()},
                    **{k + "_calls": flatten(v) for k, v in calls.items()},
                    **{k: v for k, v in feedback_cost.items()},
                }

                for col in self.rec_cols:
                    row[col] = (
                        datetime.fromtimestamp(_rec.ts).isoformat()
                        if col == "ts"
                        else getattr(_rec, col)
                    )

                yield row

            except Exception as e:
                # Handling unexpected errors, possibly due to database issues.
                print(
                    "Error encountered while attempting to retrieve feedback results. "
                    "This issue may stem from a corrupted database."
                )
                print(f"Error details: {e}")


def flatten(nested: Iterable[Iterable[Any]]) -> List[Any]:
    def _flatten(
        _nested: Iterable[Iterable[Any]],
    ) -> Generator[Any, None, None]:
        for iterable in _nested:
            yield from iterable

    return list(_flatten(nested))



"""
# Serialization of Python objects

In order to serialize (and optionally deserialize) Python entities while still
being able to inspect them in their serialized form, we employ several storage
classes that mimic basic Python entities:

| Serializable representation | Python entity |
| --- | --- |
| Class    | (python) class    |
| Module   | (python) module   |
| Obj      | (python) object   |
| Function | (python) function |
| Method   | (python) method   |

"""

from __future__ import annotations

import functools
import importlib
import inspect
import logging
from pprint import PrettyPrinter
from types import ModuleType
from typing import Any, Callable, Dict, Optional, Sequence, Tuple
import warnings

import pydantic
from trulens.core.utils import constants as constant_utils
from trulens.core.utils import python as python_utils
from trulens.core.utils import serial as serial_utils

logger = logging.getLogger(__name__)
pp = PrettyPrinter()


def is_noserio(obj: Any) -> bool:
    """
    Determines whether the given json object represents some non-serializable
    object. See `noserio`.
    """
    return isinstance(obj, dict) and constant_utils.NOSERIO in obj


def noserio(obj: Any, **extra: Dict) -> Dict:
    """
    Create a json structure to represent a non-serializable object. Any
    additional keyword arguments are included.
    """

    inner = Obj.of_object(obj).model_dump()
    inner.update(extra)

    if isinstance(obj, Sequence):
        inner["len"] = len(obj)

    return {constant_utils.NOSERIO: inner}


# TODO: rename as functionality optionally produces JSONLike .
def safe_getattr(obj: Any, k: str, get_prop: bool = True) -> Any:
    """
    Try to get the attribute `k` of the given object. If it’s a @property or
    functools.cached_property (including OpenAI’s cached_property shim), this
    will invoke the descriptor to compute and return the value.

    If `get_prop` is False, will raise ValueError for any property/descriptor.
    On exceptions, returns a dict with an ERROR key.
    """
    # Fetch the descriptor or raw attribute without invoking it
    v = inspect.getattr_static(obj, k)

    # Detect both plain @property and stdlib cached_property
    is_prop = isinstance(v, property)
    is_std_cached = isinstance(v, functools.cached_property)

    if is_prop or is_std_cached:
        if not get_prop:
            raise ValueError(f"{k} is a property/descriptor")
        try:
            # Invoke the descriptor protocol to compute (and cache) the value
            return v.__get__(obj, type(obj))
        except Exception as e:
            return {constant_utils.ERROR: Obj.of_object(e)}

    # Not a property/descriptor we care about—return the raw value
    return v


def clean_attributes(obj, include_props: bool = False) -> Dict[str, Any]:
    """
    Determine which attributes of the given object should be enumerated for
    storage and/or display in UI. Returns a dict of those attributes and their
    values.

    For enumerating contents of objects that do not support utility classes like
    pydantic, we use this method to guess what should be enumerated when
    serializing/displaying.

    If `include_props` is True, will produce attributes which are properties;
    otherwise those will be excluded.
    """

    keys = dir(obj)

    ret = {}

    for k in keys:
        if k.startswith("__"):
            # These are typically very internal components not meant to be
            # exposed beyond immediate definitions. Ignoring these.
            continue

        if include_props and k.startswith("_") and k[1:] in keys:
            # Objects often have properties named `name` with their values
            # coming from `_name`. This check lets avoid including both the
            # property and the value.
            continue

        try:
            v = safe_getattr(obj, k, get_prop=include_props)

            if isinstance(v, python_utils.OpaqueWrapper):
                # Don't expose the contents of opaque wrappers.
                continue

            ret[k] = v

        except Exception as e:
            logger.debug(str(e))
            logger.debug(str(e.__traceback__))

    return ret


class Module(serial_utils.SerialModel):
    package_name: Optional[str] = None  # some modules are not in a package
    module_name: str

    def of_module(mod: ModuleType, loadable: bool = False) -> "Module":
        if loadable and mod.__name__ == "__main__":
            # running in notebook
            raise ImportError(f"Module {mod} is not importable.")

        return Module(package_name=mod.__package__, module_name=mod.__name__)

    def of_module_name(module_name: str, loadable: bool = False) -> "Module":
        if loadable and module_name == "__main__":
            # running in notebook
            raise ImportError(f"Module {module_name} is not importable.")

        try:
            mod = importlib.import_module(module_name)
            package_name = mod.__package__
            return Module(package_name=package_name, module_name=module_name)

        except ImportError:
            return Module(package_name=None, module_name=module_name)

    def load(self) -> ModuleType:
        return importlib.import_module(
            self.module_name, package=self.package_name
        )


class Class(serial_utils.SerialModel):
    """
    A Python class. Should be enough to deserialize the constructor. Also
    includes bases so that we can query subtyping relationships without
    deserializing the class first.
    """

    name: str

    module: Module

    bases: Optional[Sequence[Class]] = None

    def __repr__(self):
        return self.module.module_name + "." + self.name

    def __str__(self):
        return f"{self.name}({self.module.module_name if self.module is not None else 'no module'})"

    def base_class(self) -> Class:
        """
        Get the deepest base class in the same module as this class.
        """
        if self.bases is None:
            return self

        module_name = self.module.module_name

        for base in self.bases[::-1]:
            if base.module.module_name == module_name:
                return base

        return self

    def _check_importable(self):
        try:
            self.load()
        except Exception as e:
            print(e)
            raise ImportError(
                f"Class {self} is not importable. "
                "If you are defining custom feedback function implementations, make sure they can be imported by Python scripts. "
                "If you defined a function in a notebook, it will not be importable."
            )

    @staticmethod
    def of_class(
        cls: type, with_bases: bool = False, loadable: bool = False
    ) -> "Class":
        ret = Class(
            name=cls.__name__,
            module=Module.of_module_name(object_module(cls), loadable=loadable),
            bases=list(map(lambda base: Class.of_class(cls=base), cls.__mro__))
            if with_bases
            else None,
        )

        if loadable:
            if "<locals>" in repr(
                cls
            ):  # TODO: figure out a better way to check this
                raise ImportError(f"Class {cls} is not globally importable.")

            ret._check_importable()

        return ret

    @staticmethod
    def of_object(
        obj: object, with_bases: bool = False, loadable: bool = False
    ):
        return Class.of_class(
            cls=obj.__class__, with_bases=with_bases, loadable=loadable
        )

    @staticmethod
    def of_class_info(json: serial_utils.JSON):
        assert constant_utils.CLASS_INFO in json, "Class info not in json."
        return Class.model_validate(json[constant_utils.CLASS_INFO])

    def load(self) -> type:  # class
        try:
            mod = self.module.load()
            return getattr(mod, self.name)

        except Exception as e:
            raise RuntimeError(f"Could not load class {self} because {e}.")

    def noserio_issubclass(self, class_name: str, module_name: str):
        bases = self.bases

        assert (
            bases is not None
        ), "Cannot do subclass check without bases. Serialize me with `Class.of_class(with_bases=True ...)`."

        for base in bases:
            if (
                base.name == class_name
                and base.module.module_name == module_name
            ):
                return True

        return False


Class.model_rebuild()


# inspect.signature does not work on builtin type constructors but they are used
# like this method. Use it to create a signature of a builtin constructor.
def builtin_init_dummy(self, *args, **kwargs):
    pass


builtin_init_sig = inspect.signature(builtin_init_dummy)


def _safe_init_sig(cls):
    """
    Get the signature of the constructor method of the given class `cls`. If it is
    a builtin class, this typically raises an exception in which case we return
    a generic signature that seems typical of builtin constructors.
    """

    try:
        return inspect.signature(cls)
    except Exception:
        return builtin_init_sig


class Obj(serial_utils.SerialModel):
    # TODO: refactor this into something like WithClassInfo, perhaps
    # WithObjectInfo, and store required constructor inputs as attributes with
    # potentially a placeholder for additional arguments that are not
    # attributes, under a special key like "__tru_object_info".
    """
    An object that may or may not be loadable from its serialized form. Do not
    use for base types that don't have a class. Loadable if `init_bindings` is
    not None.
    """

    cls: Class

    # From id(obj), identifies memory location of a python object. Use this for
    # handling loops in JSON objects.
    id: int

    # Loadable
    init_bindings: Optional[Bindings] = None

    @staticmethod
    def of_object(
        obj: object, cls: Optional[type] = None, loadable: bool = False
    ) -> Obj:
        if cls is None:
            cls = obj.__class__

        bindings = None

        if loadable:
            # Constructor arguments for some common types.
            if isinstance(obj, pydantic.BaseModel):
                # NOTE: avoids circular import:

                init_args = ()
                init_kwargs = obj.model_dump()
                # init_kwargs = json_utils.jsonify(obj)

            elif isinstance(obj, Exception):
                init_args = obj.args
                init_kwargs = {}

            else:
                # For unknown types, check if the constructor for cls expect
                # arguments and fail if so as we don't know where to get them. If
                # there are none, create empty init bindings.

                sig = _safe_init_sig(cls)
                if len(sig.parameters) > 0:
                    raise RuntimeError(
                        f"Do not know how to get constructor arguments for object of type {cls.__name__}. "
                        f"If you are defining a custom feedback function, define its implementation as a function or a method of a Provider subclass."
                    )

                init_args = ()
                init_kwargs = {}

            # TODO: dataclasses
            # TODO: dataclasses_json

            # NOTE: Something related to pydantic models incorrectly sets signature
            # of cls so we need to check cls.__call__ instead.
            # TODO: app serialization
            # if isinstance(cls, type):
            #    sig = _safe_init_sig(cls)
            # else:

            sig = _safe_init_sig(cls.__call__)

            b = sig.bind(*init_args, **init_kwargs)
            bindings = Bindings.of_bound_arguments(b)

        cls_serial = Class.of_class(cls)

        if loadable:
            cls_serial._check_importable()

        return Obj(cls=cls_serial, id=id(obj), init_bindings=bindings)

    def load(self) -> object:
        if self.init_bindings is None:
            raise RuntimeError(
                "Cannot load object unless `init_bindings` is set."
            )

        cls = self.cls.load()

        if issubclass(cls, pydantic.BaseModel):
            # For pydantic Models, use model_validate to reconstruct object:
            return cls.model_validate(self.init_bindings.kwargs)

        else:
            sig = _safe_init_sig(cls)

            if (
                constant_utils.CLASS_INFO in sig.parameters
                and constant_utils.CLASS_INFO not in self.init_bindings.kwargs
            ):
                extra_kwargs = {constant_utils.CLASS_INFO: self.cls}
            else:
                extra_kwargs = {}

            try:
                bindings = self.init_bindings.load(
                    sig, extra_kwargs=extra_kwargs
                )

            except Exception as e:
                msg = "Error binding constructor args for object:\n"
                msg += str(e) + "\n"
                msg += f"\tobj={self}\n"
                msg += f"\targs={self.init_bindings.args}\n"
                msg += f"\tkwargs={self.init_bindings.kwargs}\n"
                raise type(e)(msg)

            return cls(*bindings.args, **bindings.kwargs)


def _self_arg(bindings: inspect.BoundArguments) -> Optional[str]:
    """Guess whether the given bindings have a "self" argument and return its
    name.

    This guesses that first arg that contains "self" is a self argument.
    """

    if len(bindings.arguments) == 0:
        return None

    firstarg = next(iter(bindings.arguments))

    if "self" in firstarg:
        return firstarg

    return None


class Bindings(serial_utils.SerialModel):
    args: Tuple
    kwargs: Dict[str, Any]

    @staticmethod
    def of_bound_arguments(
        b: inspect.BoundArguments,
        skip_self: bool = True,
        arguments_only: bool = False,
    ) -> Bindings:
        """Populate Bindings from inspect.BoundArguments.

        Args:
            b: BoundArguments to populate from.

            skip_self: If True, skip the first argument if it is named "self".

            arguments_only: If True, only populate kwargs from arguments. This
                includes the same arguments as otherwise except it provides all of
                them by name even if they were bound by position.
        """

        firstarg: Optional[str] = _self_arg(b)

        if arguments_only:
            return Bindings(
                args=(),
                kwargs={
                    k: v
                    for k, v in b.arguments.items()
                    if (not skip_self or k != firstarg)
                },
            )

        if skip_self:
            if firstarg is not None:
                args = b.args[1:]
            else:
                args = b.args
        else:
            args = b.args

        return Bindings(args=args, kwargs=b.kwargs)

    def _handle_providers_load(self):
        # HACK004: A Hack: reason below
        # This was introduced with the feedback functions Groundedness and GroundTruthAgreement.

        # `summarize_provider` explanation:
        ## The use case is a Serialized feedback, with attribute that needs instantiation
        ## But should not be a user supplied input kwarg.
        # `groundedness_provider` and `provider` explanation
        ## The rest of the providers need to be instantiated, but are currently in circular dependency if done from util.py

        if "summarize_provider" in self.kwargs:
            del self.kwargs["summarize_provider"]
        if "groundedness_provider" in self.kwargs:
            del self.kwargs["groundedness_provider"]
        if "provider" in self.kwargs:
            del self.kwargs["provider"]

    def load(self, sig: inspect.Signature, extra_args=(), extra_kwargs={}):
        # Disabling this hack as we now have different providers that may need
        # to be selected from (i.e. OpenAI vs AzureOpenAI).

        # self._handle_providers_load()

        return sig.bind(
            *(self.args + extra_args), **self.kwargs, **extra_kwargs
        )


class FunctionOrMethod(serial_utils.SerialModel):
    @classmethod
    def model_validate(cls, obj, **kwargs):
        if isinstance(obj, Dict):
            if "obj" in obj:
                return super(cls, Method).model_validate(obj=obj, **kwargs)
            elif "cls" in obj:
                return super(cls, Function).model_validate(obj=obj, **kwargs)
            else:
                raise ValueError(
                    f"Cannot tell what type of callable this encodes: {obj}"
                )
        else:
            return super().model_validate(obj)

    @staticmethod
    def of_callable(c: Callable, loadable: bool = False) -> "FunctionOrMethod":
        """Serialize the given callable.

        If `loadable` is set, tries to add enough info for the callable to be
        deserialized.
        """

        if inspect.ismethod(c):
            if not hasattr(c, "__self__"):
                raise ValueError(
                    f"Method {c} does not have a self object. Cannot serialize."
                )
            self = c.__self__
            return Method.of_method(c, obj=self, loadable=loadable)

        else:
            return Function.of_function(c, loadable=loadable)

    def load(self) -> Callable:
        raise NotImplementedError()


class Method(FunctionOrMethod):
    """
    A Python method. A method belongs to some class in some module and must have
    a pre-bound self object. The location of the method is encoded in `obj`
    alongside self. If obj is Obj with init_bindings, this method should be
    deserializable.
    """

    obj: Obj
    name: str

    @staticmethod
    def of_method(
        meth: Callable,
        cls: Optional[type] = None,
        obj: Optional[object] = None,
        loadable: bool = False,
    ) -> "Method":
        if obj is None:
            assert inspect.ismethod(
                meth
            ), f"Expected a method (maybe it is a function?): {meth}"
            obj = meth.__self__

        if cls is None:
            if isinstance(cls, type):
                # classmethod, self is a type
                cls = obj
            else:
                # normal method, self is instance of cls
                cls = obj.__class__

        obj_model = Obj.of_object(obj, cls=cls, loadable=loadable)

        return Method(obj=obj_model, name=meth.__name__)

    def load(self) -> Callable:
        obj = self.obj.load()
        return getattr(obj, self.name)


def object_module(obj):
    if python_utils.safe_hasattr(obj, "__module__"):
        return getattr(obj, "__module__")
    else:
        return "builtins"


class Function(FunctionOrMethod):
    """
    A Python function. Could be a static method inside a class (not instance of
    the class).
    """

    module: Module

    # For static methods in a class which we view as functions, not yet
    # supported:
    cls: Optional[Class]

    name: str

    @staticmethod
    def of_function(
        func: Callable,
        module: Optional[ModuleType] = None,
        cls: Optional[type] = None,
        loadable: bool = False,
    ) -> "Function":  # actually: class
        if module is None:
            module = Module.of_module_name(
                object_module(func), loadable=loadable
            )

        if cls is not None:
            cls = Class.of_class(cls, loadable=loadable)

        if isinstance(func, functools.partial):
            f_name = func.func.__name__
        else:
            f_name = func.__name__

        return Function(cls=cls, module=module, name=f_name)

    def load(self) -> Callable:
        if self.cls is not None:
            # TODO: static/class methods work in progress

            cls = self.cls.load()  # does not create object instance
            return getattr(cls, self.name)  # lookup static/class method

        else:
            mod = self.module.load()
            try:
                return getattr(mod, self.name)  # function not inside a class
            except Exception:
                raise ImportError(
                    f"Function {self} is not importable. "
                    "If you are defining custom feedback function implementations, make sure they can be imported by Python scripts. "
                    "If you defined a function in a notebook, it will not be importable."
                )


class WithClassInfo(pydantic.BaseModel):
    """Mixin to track class information to aid in querying serialized components
    without having to load them.
    """

    tru_class_info: Class
    """Class information of this pydantic object for use in deserialization.

    Using this odd key to not pollute attribute names in whatever class we mix
    this into. Should be the same as CLASS_INFO.
    """

    # NOTE(piotrm): HACK005: for some reason, model_validate is not called for
    # nested models but the method decorated as such below is called. We use
    # this to load an object which includes our class information instead of
    # using pydantic for this loading as it would always load the object as per
    # its declared field. For example, `Provider` includes `endpoint: Endpoint`
    # but we want to load one of the `Endpoint` subclasses. We add the subclass
    # information using `WithClassInfo` meaning we can then use this method
    # below to load the subclass. Pydantic would only give us `Endpoint`, the
    # parent class.
    @pydantic.model_validator(mode="before")
    @staticmethod
    def load(obj, *args, **kwargs):
        """Deserialize/load this object using the class information in
        tru_class_info to lookup the actual class that will do the deserialization."""

        if not isinstance(obj, dict):
            return obj

        if constant_utils.CLASS_INFO not in obj:
            raise ValueError("No class info present in object.")

        clsinfo = Class.model_validate(obj[constant_utils.CLASS_INFO])
        try:
            # If class cannot be loaded, usually because it is not importable,
            # return obj as is.
            cls = clsinfo.load()
        except RuntimeError:
            return obj

        # TODO: We shouldn't be doing a lot of these pydantic details but could
        # not find how to integrate with existing pydantic functionality. Please
        # figure it out.
        validated = {}

        for k, finfo in cls.model_fields.items():
            typ = finfo.annotation
            val = finfo.get_default(call_default_factory=True)

            if k in obj:
                val = obj[k]

            try:
                if (
                    (isinstance(val, dict))
                    and (constant_utils.CLASS_INFO in val)
                    and inspect.isclass(typ)
                    and python_utils.safe_issubclass(typ, WithClassInfo)
                ):
                    subcls = Class.model_validate(
                        val[constant_utils.CLASS_INFO]
                    ).load()

                    val = subcls.model_validate(val)
            except Exception:
                pass

            validated[k] = val

        # Note that the rest of the validation/conversions for things which are
        # not serialized WithClassInfo will be done by pydantic after we return
        # this:
        return validated

    def __init__(
        self,
        *args,
        class_info: Optional[Class] = None,
        obj: Optional[object] = None,
        cls: Optional[type] = None,
        **kwargs,
    ):
        if obj is not None:
            warnings.warn(
                "`obj` does not need to be provided to WithClassInfo any more",
                DeprecationWarning,
            )

        if obj is None:
            obj = self

        if obj is not None:
            cls = type(obj)

        if class_info is None:
            assert (
                cls is not None
            ), "Either `class_info`, `obj` or `cls` need to be specified."
            class_info = Class.of_class(cls, with_bases=True)

        kwargs[constant_utils.CLASS_INFO] = class_info

        super().__init__(*args, **kwargs)

    @staticmethod
    def get_class(obj_json: Dict):
        return Class.model_validate(obj_json[constant_utils.CLASS_INFO]).load()

    @staticmethod
    def of_object(obj: object):
        return WithClassInfo(class_info=Class.of_class(obj.__class__))

    @staticmethod
    def of_class(cls: type):  # class
        return WithClassInfo(class_info=Class.of_class(cls))

    @classmethod
    def model_validate(cls, *args, **kwargs) -> Any:
        # Note: This is here only so we can provide a pointer and some
        # instructions to render into the docs.
        """
        Deserialized a jsonized version of the app into the instance of the
        class it was serialized from.

        Note:
            This process uses extra information stored in the jsonized object
            and handled by [WithClassInfo][trulens.core.utils.pyschema.WithClassInfo].
        """

        return super().model_validate(*args, **kwargs)


# HACK013:
Module.model_rebuild()
Class.model_rebuild()
Obj.model_rebuild()
Bindings.model_rebuild()
FunctionOrMethod.model_rebuild()
Function.model_rebuild()
Method.model_rebuild()
WithClassInfo.model_rebuild()



"""Update app_json with app_id from apps table.

Revision ID: 9
Revises: 8
Create Date: 2024-08-16 12:46:49.510690
"""

import json

from alembic import op
from sqlalchemy.orm.session import Session
from trulens.core.database.orm import make_orm_for_prefix

try:
    from tqdm import tqdm
except ImportError:
    tqdm = None

# revision identifiers, used by Alembic.
revision = "9"
down_revision = "8"
branch_labels = None
depends_on = "7"


def upgrade(config) -> None:
    prefix = config.get_main_option("trulens.table_prefix")

    if prefix is None:
        raise RuntimeError("trulens.table_prefix is not set")

    # ### begin Alembic commands ###
    with Session(bind=op.get_bind()) as session:
        orm = make_orm_for_prefix(table_prefix=prefix)

        apps = session.query(orm.AppDefinition).all()
        if tqdm is not None:
            apps = tqdm(apps, desc="Updating app_json in apps table")
        for app in apps:
            if app.app_json is None:
                continue
            app_json = json.loads(app.app_json)
            app_json["app_id"] = app.app_id
            if "app_name" not in app_json:
                app_json["app_name"] = app.app_name
            if "app_version" not in app_json:
                app_json["app_version"] = app.app_version
            app.app_json = json.dumps(app_json)
        session.commit()
    # ### end Alembic commands ###


def downgrade(config) -> None:
    prefix = config.get_main_option("trulens.table_prefix")

    if prefix is None:
        raise RuntimeError("trulens.table_prefix is not set")

    # ### begin Alembic commands ###

    with Session(bind=op.get_bind()) as session:
        orm = make_orm_for_prefix(table_prefix=prefix)

        apps = session.query(orm.AppDefinition).all()
        for app in apps:
            if app.app_json is None:
                continue
            app_json = json.loads(app.app_json)
            app_json["app_id"] = app.app_version
            if "app_name" in app_json:
                del app_json["app_name"]
            if "app_version" in app_json:
                del app_json["app_version"]
            app.app_json = json.dumps(app_json)
        session.commit()
    # ### end Alembic commands ###



"""Serialization utilities.

TODO: Lens class: can we store just the python AST instead of building up our
own "Step" classes to hold the same data? We are already using AST for parsing.
"""

from __future__ import annotations

import ast
from ast import dump
from ast import parse
import base64
from contextvars import ContextVar
from copy import copy
import logging
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Generic,
    Hashable,
    Iterable,
    List,
    Optional,
    Sequence,
    Set,
    Sized,
    Tuple,
    TypeVar,
    Union,
)

from munch import Munch as Bunch
import pydantic
from pydantic import GetCoreSchemaHandler
from pydantic.v1 import BaseModel as v1BaseModel
from pydantic_core import CoreSchema
from pydantic_core import core_schema
import rich.repr
from trulens.core.utils import containers as container_utils
from trulens.core.utils import python as python_utils

logger = logging.getLogger(__name__)

T = TypeVar("T")

JSON_BASES: Tuple[type, ...] = (str, int, float, bytes, type(None))
"""
Tuple of JSON-able base types.

Can be used in `isinstance` checks.
"""

JSON_BASES_T = Union[str, int, float, bytes, None]
"""
Alias for JSON-able base types.
"""

JSON = Union[JSON_BASES_T, Sequence[Any], Dict[str, Any]]
"""Alias for (non-strict) JSON-able data (`Any` = `JSON`).

If used with type argument, that argument indicates what the JSON represents and
can be desererialized into.

Formal JSON must be a `dict` at the root but non-strict here means that the root
can be a basic type or a sequence as well.
"""


def is_strict_json(obj: Any) -> bool:
    """Determine if the given object is JSON-able, strictly.

    Strict JSON starts as a dictionary at the root.
    """

    return isinstance(obj, dict)


def is_json(obj: Any) -> bool:
    """Determine if the given object is JSON-able."""

    return isinstance(obj, (*JSON_BASES, Sequence, Dict))


JSON_STRICT = Dict[str, JSON]
"""
Alias for (strictly) JSON-able data.

Python object that is directly mappable to JSON.
"""


class JSONized(dict, Generic[T]):  # really JSON_STRICT
    """JSON-encoded data the can be deserialized into a given type `T`.

    This class is meant only for type annotations. Any
    serialization/deserialization logic is handled by different classes, usually
    subclasses of `pydantic.BaseModel`.
    """

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        """Make pydantic treat this class same as a `dict`."""
        return handler(core_schema.dict_schema())


# NOTE1: Lens, a container for selector/accessors/setters of data stored in a
# json structure. Cannot make abstract since pydantic will try to initialize it.
class Step(pydantic.BaseModel, Hashable):
    """
    A step in a selection path.
    """

    def __hash__(self) -> int:
        raise TypeError(f"Should never be called, self={self.model_dump()}")

    @classmethod
    def model_validate(cls, obj: Union[Step, Dict], **kwargs) -> Step:
        if isinstance(obj, Step):
            return super().model_validate(obj, **kwargs)

        elif isinstance(obj, dict):
            ATTRIBUTE_TYPE_MAP = {
                "item": GetItem,
                "index": GetIndex,
                "attribute": GetAttribute,
                "item_or_attribute": GetItemOrAttribute,
                "start": GetSlice,
                "stop": GetSlice,
                "step": GetSlice,
                "items": GetItems,
                "indices": GetIndices,
                "collect": Collect,
            }

            a = next(iter(obj.keys()))
            if a in ATTRIBUTE_TYPE_MAP:
                return ATTRIBUTE_TYPE_MAP[a](**obj)

        raise RuntimeError(
            f"Do not know how to interpret {obj} as a `Lens` `Step`."
        )

    # @abc.abstractmethod # NOTE1
    def get(self, obj: Any) -> Iterable[Any]:
        """
        Get the element of `obj`, indexed by `self`.
        """
        raise NotImplementedError()

    # @abc.abstractmethod # NOTE1
    def set(self, obj: Any, val: Any) -> Any:
        """
        Set the value(s) indicated by self in `obj` to value `val`.
        """
        raise NotImplementedError()

    def get_sole_item(self, obj: Any) -> Any:
        all_objects = list(self.get(obj))

        if len(all_objects) != 1:
            raise ValueError(
                f"Step {self} did not address exactly a single object."
            )

        return all_objects[0]


class Collect(Step):
    # Need something for `Step.validate` to tell that it is looking at Collect.
    collect: None = None

    # Hashable requirement.
    def __hash__(self):
        return hash("collect")

    # Step requirement.
    def get(self, obj: Any) -> Iterable[List[Any]]:
        # Needs to be handled in Lens class itself.
        raise NotImplementedError()

    # Step requirement
    def set(self, obj: Any, val: Any) -> Any:
        raise NotImplementedError()

    def __repr__(self) -> str:
        return ".collect()"


class StepItemOrAttribute(Step):
    # NOTE1
    def get_item_or_attribute(self):
        raise NotImplementedError()


class GetAttribute(StepItemOrAttribute):
    """An attribute lookup step as in `someobject.someattribute`."""

    attribute: str

    # Hashable requirement.
    def __hash__(self) -> int:
        return hash(self.attribute)

    # StepItemOrAttribute requirement
    def get_item_or_attribute(self) -> str:
        return self.attribute

    # Step requirement
    def get(self, obj: Any) -> Iterable[Any]:
        if hasattr(obj, self.attribute):
            yield getattr(obj, self.attribute)
        else:
            raise ValueError(
                f"Object {obj} does not have attribute: {self.attribute}"
            )

    # Step requirement
    def set(self, obj: Any, val: Any) -> Any:
        if obj is None:
            obj = Bunch()

        # might cause issues
        obj = copy(obj)

        if hasattr(obj, self.attribute):
            setattr(obj, self.attribute, val)
            return obj
        else:
            # might fail
            setattr(obj, self.attribute, val)
            return obj

    def __repr__(self) -> str:
        return f".{self.attribute}"


class GetIndex(Step):
    """An index lookup step as in `someobject[5]`."""

    index: int

    # Hashable requirement
    def __hash__(self) -> int:
        return hash(self.index)

    # Step requirement
    def get(self, obj: Sequence[T]) -> Iterable[T]:
        if isinstance(obj, Sequence):
            if len(obj) > self.index:
                yield obj[self.index]
            else:
                raise IndexError(f"Index out of bounds: {self.index}")
        else:
            raise ValueError(f"Object {obj} is not a sequence.")

    # Step requirement
    def set(self, obj: Any, val: Any) -> Any:
        if obj is None:
            obj = []

        assert isinstance(obj, Sequence), "Sequence expected."

        # copy
        obj = list(obj)

        if self.index >= 0:
            while len(obj) <= self.index:
                obj.append(None)

        obj[self.index] = val
        return obj

    def __repr__(self) -> str:
        return f"[{self.index}]"


class GetItem(StepItemOrAttribute):
    """An item lookup step as in `someobject["somestring"]`."""

    item: str

    # Hashable requirement
    def __hash__(self) -> int:
        return hash(self.item)

    # StepItemOrAttribute requirement
    def get_item_or_attribute(self) -> str:
        return self.item

    # Step requirement
    def get(self, obj: Dict[str, T]) -> Iterable[T]:
        if isinstance(obj, Dict):
            if self.item in obj:
                yield obj[self.item]
            else:
                raise KeyError(f"Key not in dictionary: {self.item}")
        else:
            raise ValueError(f"Object {obj} is not a dictionary.")

    # Step requirement
    def set(self, obj: Any, val: Any) -> Any:
        if obj is None:
            obj = dict()

        assert isinstance(obj, Dict), "Dictionary expected."

        # copy
        obj = {k: v for k, v in obj.items()}

        obj[self.item] = val
        return obj

    def __repr__(self) -> str:
        return f"[{repr(self.item)}]"


class GetItemOrAttribute(StepItemOrAttribute):
    """A step in a path lens that selects an item or an attribute.

    Note:
        _TruLens_ allows looking up elements within sequences if the subelements
        have the item or attribute. We issue warning if this is ambiguous (looking
        up in a sequence of more than 1 element).
    """

    item_or_attribute: str  # distinct from "item" for deserialization

    # Hashable requirement
    def __hash__(self) -> int:
        return hash(self.item_or_attribute)

    # StepItemOrAttribute requirement
    def get_item_or_attribute(self) -> str:
        return self.item_or_attribute

    # Step requirement
    def get(self, obj: Dict[str, T]) -> Iterable[T]:
        # Special handling of sequences. See NOTE above.

        if isinstance(obj, Sequence) and not isinstance(obj, str):
            if len(obj) == 1:
                yield from self.get(obj=obj[0])
            elif len(obj) == 0:
                raise ValueError(
                    f"Object not a dictionary or sequence of dictionaries: {obj}."
                )
            else:  # len(obj) > 1
                logger.warning(
                    "Object (of type %s is a sequence containing more than one dictionary. "
                    "Lookup by item or attribute `%s` is ambiguous. "
                    "Use a lookup by index(es) or slice first to disambiguate.",
                    type(obj).__name__,
                    self.item_or_attribute,
                )
                for sub_obj in obj:
                    try:
                        yield from self.get(obj=sub_obj)
                    except Exception:
                        pass

        # Otherwise handle a dict or object with the named attribute.
        elif isinstance(obj, Dict):
            if self.item_or_attribute in obj:
                yield obj[self.item_or_attribute]
            else:
                raise KeyError(
                    f"Key not in dictionary: {self.item_or_attribute}"
                )
        else:
            if hasattr(obj, self.item_or_attribute):
                yield getattr(obj, self.item_or_attribute)
            else:
                raise ValueError(
                    f"Object {repr(obj)} of type {type(obj)} does not have item or attribute {repr(self.item_or_attribute)}."
                )

    # Step requirement
    def set(self, obj: Any, val: Any) -> Any:
        if obj is None:
            obj = dict()

        if isinstance(obj, Dict) and not isinstance(obj, Bunch):
            # Bunch claims to be a Dict.
            # copy
            obj = {k: v for k, v in obj.items()}
            obj[self.item_or_attribute] = val
        else:
            obj = copy(obj)  # might cause issues
            setattr(obj, self.item_or_attribute, val)

        return obj

    def __repr__(self) -> str:
        return f".{self.item_or_attribute}"


class GetSlice(Step):
    start: Optional[int]
    stop: Optional[int]
    step: Optional[int]

    # Hashable requirement
    def __hash__(self) -> int:
        return hash((self.start, self.stop, self.step))

    # Step requirement
    def get(self, obj: Sequence[T]) -> Iterable[T]:
        if isinstance(obj, Sequence):
            lower, upper, step = slice(
                self.start, self.stop, self.step
            ).indices(len(obj))
            for i in range(lower, upper, step):
                yield obj[i]
        else:
            raise ValueError("Object is not a sequence.")

    # Step requirement
    def set(self, obj: Any, val: Any) -> Any:
        if obj is None:
            obj = []

        assert isinstance(obj, Sequence), "Sequence expected."

        lower, upper, step = slice(self.start, self.stop, self.step).indices(
            len(obj)
        )

        # copy
        obj = list(obj)

        for i in range(lower, upper, step):
            obj[i] = val

        return obj

    def __repr__(self) -> str:
        pieces = ":".join([
            "" if p is None else str(p)
            for p in (self.start, self.stop, self.step)
        ])
        if pieces == "::":
            pieces = ":"

        return f"[{pieces}]"


class GetIndices(Step):
    indices: Tuple[int, ...]

    # Hashable requirement
    def __hash__(self) -> int:
        return hash(self.indices)

    def __init__(self, indices: Iterable[int]):
        super().__init__(indices=tuple(indices))

    # Step requirement
    def get(self, obj: Sequence[T]) -> Iterable[T]:
        if isinstance(obj, Sequence):
            for i in self.indices:
                yield obj[i]
        else:
            raise ValueError("Object is not a sequence.")

    # Step requirement
    def set(self, obj: Any, val: Any) -> Any:
        if obj is None:
            obj = []

        assert isinstance(obj, Sequence), "Sequence expected."

        # copy
        obj = list(obj)

        for i in self.indices:
            if i >= 0:
                while len(obj) <= i:
                    obj.append(None)

            obj[i] = val

        return obj

    def __repr__(self) -> str:
        return f"[{','.join(map(str, self.indices))}]"


class GetItems(Step):
    items: Tuple[str, ...]

    # Hashable requirement
    def __hash__(self) -> str:
        return hash(self.items)

    def __init__(self, items: Iterable[str]):
        super().__init__(items=tuple(items))

    # Step requirement
    def get(self, obj: Dict[str, T]) -> Iterable[T]:
        if isinstance(obj, Dict):
            for i in self.items:
                yield obj[i]
        else:
            raise ValueError("Object is not a dictionary.")

    # Step requirement
    def set(self, obj: Any, val: Any) -> Any:
        if obj is None:
            obj = dict()

        assert isinstance(obj, Dict), "Dictionary expected."

        # copy
        obj = {k: v for k, v in obj.items()}

        for i in self.items:
            obj[i] = val

        return obj

    def __repr__(self) -> str:
        return "[" + (",".join(f"'{i}'" for i in self.items)) + "]"


def model_dump(obj: Union[pydantic.BaseModel, v1BaseModel]) -> dict:
    """
    Return the dict/model_dump of the given pydantic instance regardless of it
    being v2 or v1.
    """

    if isinstance(obj, pydantic.BaseModel):
        return obj.model_dump()
    elif isinstance(obj, v1BaseModel):
        return obj.dict()
    else:
        raise ValueError("Not a pydantic.BaseModel.")


class SerialModel(pydantic.BaseModel):
    """
    Trulens-specific additions on top of pydantic models. Includes utilities to
    help serialization mostly.
    """

    formatted_objects: ClassVar[ContextVar[Set[int]]] = ContextVar(
        "formatted_objects", default=set()
    )

    def _enter_formatting(self) -> Tuple[Optional[Any], bool]:
        """Begin formatting this object, tracking circular refs.

        Returns:
            A tuple of ``(token, is_circular)``.  When
            *is_circular* is ``True`` this object has already been
            seen in the current formatting chain and should be
            represented with a short placeholder.  *token* must be
            passed to :meth:`_exit_formatting` by the **root**
            formatter when it finishes.
        """

        tok = None
        if SerialModel.formatted_objects.get(None) is None:
            tok = SerialModel.formatted_objects.set(set())

        formatted_objects = SerialModel.formatted_objects.get()

        if formatted_objects is None:
            formatted_objects = set()

        if id(self) in formatted_objects:
            return tok, True

        formatted_objects.add(id(self))
        return tok, False

    @staticmethod
    def _exit_formatting(tok: Optional[Any]) -> None:
        """Clean up after the root formatting call."""

        if tok is not None:
            SerialModel.formatted_objects.reset(tok)

    def __repr__(self) -> str:
        """Safe ``repr`` that handles circular references.

        Pydantic's default ``__repr__`` does not guard against
        circular references among model instances, which leads to
        ``RecursionError`` (see GitHub issue #1862).  This override
        uses the same ``formatted_objects`` context-variable that
        ``__rich_repr__`` uses so that already-visited objects are
        replaced with a short placeholder instead of recursing
        infinitely.
        """

        tok, is_circular = self._enter_formatting()

        if is_circular:
            SerialModel._exit_formatting(tok)
            return f"{python_utils.class_name(type(self))}" f"@0x{id(self):x}"

        try:
            fields = []
            for k, v in self.__repr_args__():
                try:
                    if k is None:
                        fields.append(repr(v))
                    else:
                        fields.append(f"{k}={v!r}")
                except RecursionError:
                    fields.append(f"{k}=..." if k is not None else "...")
            return f"{type(self).__name__}" f"({', '.join(fields)})"
        except RecursionError:
            return f"{type(self).__name__}(...)"
        finally:
            SerialModel._exit_formatting(tok)

    def __rich_repr__(self) -> rich.repr.Result:
        """Requirement for pretty printing using the rich package."""

        tok, is_circular = self._enter_formatting()

        if is_circular:
            yield (f"{python_utils.class_name(type(self))}" f"@0x{id(self):x}")
            SerialModel._exit_formatting(tok)
            return

        for k, v in self.__dict__.items():
            # This might result in recursive calls to
            # __rich_repr__ of v.
            yield k, v

        SerialModel._exit_formatting(tok)

    def model_dump_json(self, **kwargs):
        # TODO: Import at top-level and resolve circular import.
        from trulens.core.utils.json import json_str_of_obj

        return json_str_of_obj(self, **kwargs)

    def model_dump(self, **kwargs):
        # TODO: Import at top-level and resolve circular import.
        from trulens.core.utils.json import jsonify

        return jsonify(self, **kwargs)

    # NOTE(piotrm): regarding model_validate: custom deserialization is done in
    # WithClassInfo class but only for classes that mix it in.

    def update(self, **d: Dict[Any, Any]) -> SerialModel:
        for k, v in d.items():
            setattr(self, k, v)

        return self

    def replace(self, **d: Dict[Any, Any]) -> SerialModel:
        copy = self.model_copy()
        copy.update(**d)
        return copy


class SerialBytes(pydantic.BaseModel):
    # Raw data that we want to nonetheless serialize .
    data: bytes

    def dict(self):
        encoded = base64.b64encode(self.data)
        return dict(data=encoded)

    def __init__(self, data: Union[str, bytes]):
        super().__init__(data=data)

    @classmethod
    def model_validate(cls, obj, **kwargs):
        if isinstance(obj, Dict):
            encoded = obj["data"]
            if isinstance(encoded, str):
                return SerialBytes(data=base64.b64decode(encoded))
            elif isinstance(encoded, bytes):
                return SerialBytes(data=encoded)
            else:
                raise ValueError(obj)
        else:
            raise ValueError(obj)


class ParseException(Exception):
    def __init__(self, exp_string: str, exp_ast: ast.AST):
        self.exp_string = exp_string
        self.exp_ast = exp_ast

    def __str__(self) -> str:
        return (
            f"Failed to parse expression `{self.exp_string}` as a `Lens`."
            f"\nAST={dump(self.exp_ast) if self.exp_ast is not None else 'AST is None'}"
        )


class Lens(pydantic.BaseModel, Sized, Hashable):
    # Not using SerialModel as we have special handling of serialization to/from
    # strings for this class which interferes with SerialModel mechanisms.
    """
    Lenses into Python objects.

    Example:
        ```python
        path = Lens().record[5]['somekey']

        obj = ... # some object that contains a value at `obj.record[5]['somekey]`

        value_at_path = path.get(obj) # that value

        new_obj = path.set(obj, 42) # updates the value to be 42 instead
        ```

    ## `collect` and special attributes

    Some attributes hold special meaning for lenses. Attempting to access
    them will produce a special lens instead of one that looks up that
    attribute.

    Example:
        ```python
        path = Lens().record[:]

        obj = dict(record=[1, 2, 3])

        value_at_path = path.get(obj) # generates 3 items: 1, 2, 3 (not a list)

        path_collect = path.collect()

        value_at_path = path_collect.get(obj) # generates a single item, [1, 2, 3] (a list)
        ```
    """

    path: Tuple[Step, ...]

    @pydantic.model_validator(mode="wrap")
    @classmethod
    def validate_from_string(cls, obj, handler):
        # `mode="before"` validators currently cannot return something of a type
        # different than obj. Might be a pydantic oversight/bug.

        if isinstance(obj, str):
            ret = Lens.of_string(obj)
            return ret
        elif isinstance(obj, dict):
            return handler(
                dict(path=(Step.model_validate(step) for step in obj["path"]))
            )
        else:
            return handler(obj)

    @pydantic.model_serializer
    def dump_as_string(self, **kwargs):
        return str(self)

    def __init__(self, path: Optional[Iterable[Step]] = None):
        if path is None:
            path = ()

        super().__init__(path=tuple(path))

    def existing_prefix(self, obj: Any) -> Lens:
        """Get the Lens representing the longest prefix of the path that exists
        in the given object.
        """

        last_lens = Lens()
        current_lens = last_lens

        for i, step in enumerate(self.path):
            last_lens = current_lens
            current_lens = current_lens._append(step)
            if not current_lens.exists(obj):
                return last_lens

        return current_lens

    def exists(self, obj: Any) -> bool:
        """Check whether the path exists in the given object."""

        try:
            for _ in self.get(obj):
                # Check that all named values exist, not just the first one.
                pass

        except (KeyError, IndexError, ValueError):
            return False

        return True

    @staticmethod
    def of_string(s: str) -> Lens:
        """
        Convert a string representing a Python expression into a Lens.
        """

        # NOTE(piotrm): we use python parser for this which means only things
        # which are valid python expressions (with additional constraints) can
        # be converted.

        if len(s) == 0:
            return Lens()

        try:
            # NOTE: "eval" here means to parse an expression, not a statement.
            # exp = parse(f"PLACEHOLDER.{s}", mode="eval")
            exp = parse(s, mode="eval")

        except SyntaxError:
            raise ParseException(s, None)

        if not isinstance(exp, ast.Expression):
            raise ParseException(s, exp)

        exp = exp.body

        path = []

        def of_index(idx):
            if isinstance(idx, ast.Tuple):
                elts = tuple(of_index(elt) for elt in idx.elts)

                if all(isinstance(e, GetItem) for e in elts):
                    return GetItems(items=tuple(e.item for e in elts))

                elif all(isinstance(e, GetIndex) for e in elts):
                    return GetIndices(indices=tuple(e.index for e in elts))

                else:
                    raise ParseException(s, idx)

            elif isinstance(idx, ast.Constant):
                if isinstance(idx.value, str):
                    return GetItem(item=idx.value)

                elif isinstance(idx.value, int):
                    return GetIndex(index=idx.value)

                else:
                    raise ParseException(s, idx)

            elif isinstance(idx, ast.UnaryOp):
                if isinstance(idx.op, ast.USub):
                    oper = of_index(idx.operand)
                    if not isinstance(oper, GetIndex):
                        raise ParseException(s, idx)

                    return GetIndex(index=-oper.index)

            elif idx is None:
                return None

            else:
                raise ParseException(s, exp)

        while exp is not None:
            if isinstance(exp, ast.Attribute):
                attr_name = exp.attr
                path.append(GetItemOrAttribute(item_or_attribute=attr_name))
                exp = exp.value

            elif isinstance(exp, ast.Subscript):
                sub = exp.slice

                if isinstance(sub, ast.Index):
                    step = of_index(sub.value)

                    path.append(step)

                elif isinstance(sub, ast.Slice):
                    vals: Tuple[GetIndex, ...] = tuple(
                        of_index(v) for v in (sub.lower, sub.upper, sub.step)
                    )

                    if not all(
                        e is None or isinstance(e, GetIndex) for e in vals
                    ):
                        raise ParseException(s, exp)

                    vals_indices: Tuple[Union[None, int], ...] = tuple(
                        None if e is None else e.index for e in vals
                    )

                    path.append(
                        GetSlice(
                            start=vals_indices[0],
                            stop=vals_indices[1],
                            step=vals_indices[2],
                        )
                    )

                elif isinstance(sub, ast.Tuple):
                    path.append(of_index(sub))

                elif isinstance(sub, ast.Constant):
                    path.append(of_index(sub))

                else:
                    raise ParseException(s, exp)

                exp = exp.value

            elif isinstance(exp, ast.List):
                # Need this case for paths that do not start with item/attribute
                # but instead go directly to something bracketed, which is a
                # list in python syntax.

                if not all(isinstance(el, ast.Constant) for el in exp.elts):
                    raise ParseException(s, exp)

                if len(exp.elts) == 0:
                    logger.warning(
                        f"Path {s} is getting zero items/indices, it will not produce anything."
                    )
                    path.append(GetIndices(indices=()))

                elif len(exp.elts) == 1:
                    el = exp.elts[0]
                    if isinstance(el.value, int):
                        path.append(GetIndex(index=el.value))
                    elif isinstance(el.value, str):
                        path.append(GetItem(item=el.value))
                    else:
                        raise ParseException(s, exp)

                else:
                    if all(isinstance(el.value, int) for el in exp.elts):
                        path.append(
                            GetIndices(
                                indices=tuple(el.value for el in exp.elts)
                            )
                        )

                    elif all(isinstance(el.value, str) for el in exp.elts):
                        path.append(
                            GetItems(items=tuple(el.value for el in exp.elts))
                        )

                    else:
                        raise ParseException(s, exp)

                exp = None

            elif isinstance(exp, ast.Name):
                path.append(GetItemOrAttribute(item_or_attribute=exp.id))

                exp = None

            elif isinstance(exp, ast.Call):
                if not isinstance(exp.func, ast.Attribute):
                    raise ParseException(s, exp)

                funcname = exp.func.attr
                if funcname == "collect":
                    path.append(Collect())

                else:
                    raise TypeError(
                        f"`{funcname}` is not a handled call for paths."
                    )

                if len(exp.args) + len(exp.keywords) != 0:
                    logger.warning(f"args/kwargs for `{funcname}` are ignored")

                exp = exp.func.value

            else:
                raise ParseException(s, exp)

        return Lens(path=path[::-1])

    def __str__(self):
        ret = ""
        for step in self.path:
            if isinstance(step, StepItemOrAttribute) and ret == "":
                ret = step.get_item_or_attribute()
            else:
                ret += repr(step)

        return ret

    def __repr__(self):
        return "Lens()" + ("".join(map(repr, self.path)))

    # Hashable requirement
    def __hash__(self):
        return hash(self.path)

    # Sized requirement
    def __len__(self):
        return len(self.path)

    def __add__(self, other: Union[Lens, Step]) -> Lens:
        if isinstance(other, Step):
            return Lens(path=self.path + (other,))
        else:
            return Lens(path=self.path + other.path)

    def is_immediate_prefix_of(self, other: Lens):
        return self.is_prefix_of(other) and len(self.path) + 1 == len(
            other.path
        )

    def is_prefix_of(self, other: Lens):
        p = self.path
        pother = other.path

        if len(p) > len(pother):
            return False

        for s1, s2 in zip(p, pother):
            if s1 != s2:
                return False

        return True

    def set_or_append(self, obj: Any, val: Any) -> Any:
        """
        If `obj` at path `self` is None or does not exist, sets it to a list
        containing only the given `val`. If it already exists as a sequence,
        appends `val` to that sequence as a list. If it is set but not a sequence,
        error is thrown.

        """
        try:
            existing = self.get_sole_item(obj)
            if isinstance(existing, Sequence):
                return self.set(obj, list(existing) + [val])
            elif existing is None:
                return self.set(obj, [val])
            else:
                raise ValueError(
                    f"Trying to append to object which is not a list; "
                    f"is of type {type(existing).__name__} instead."
                )

        except Exception:
            return self.set(obj, [val])

    def set(self, obj: T, val: Union[Any, T]) -> T:
        """
        In `obj` at path `self` exists, change it to `val`. Otherwise create a
        spot for it with Munch objects and then set it.
        """

        if len(self.path) == 0:
            return val

        first = self.path[0]
        rest = Lens(path=self.path[1:])

        try:
            firsts = first.get(obj)
            first_obj, firsts = container_utils.iterable_peek(firsts)

        except (ValueError, IndexError, KeyError, AttributeError):
            # `first` points to an element that does not exist, use `set` to create a spot for it.
            obj = first.set(obj, None)  # will create a spot for `first`
            firsts = first.get(obj)

        for first_obj in firsts:
            obj = first.set(obj, rest.set(first_obj, val))

        return obj

    def get_sole_item(self, obj: Any) -> Any:
        all_objects = list(self.get(obj))

        assert (
            len(all_objects) == 1
        ), f"Lens {self} did not address exactly a single object."

        return all_objects[0]

    def __call__(self, *args, **kwargs):
        error_msg = (
            "Only `collect` is a valid function name in python call syntax when building lenses. "
            "Note that applying a selector/path/JSONPAth/Lens now requires to use the `get` method instead of `__call__`."
        )

        assert len(self.path) > 0, error_msg
        assert isinstance(self.path[-1], StepItemOrAttribute), error_msg

        funcname = self.path[-1].get_item_or_attribute()

        if funcname == "collect":
            return Lens(path=self.path[0:-1] + (Collect(),))

        else:
            raise TypeError(error_msg)

    def get(self, obj: Any) -> Iterable[Any]:
        if len(self.path) == 0:
            yield obj
            return

        # Need to do the recursion by breaking down steps from the end in order
        # to support `Collect`.
        last_step = self.path[-1]
        if len(self.path) == 1:
            start = Lens(path=())
        else:
            start = Lens(path=self.path[0:-1])

        start_items = start.get(obj)

        if isinstance(last_step, Collect):
            yield list(start_items)

        else:
            for start_selection in start_items:
                yield from last_step.get(start_selection)

    def _append(self, step: Step) -> Lens:
        return Lens(path=self.path + (step,))

    def __getitem__(
        self, item: Union[int, str, slice, Sequence[int], Sequence[str]]
    ) -> Lens:
        if isinstance(item, int):
            return self._append(GetIndex(index=item))
        if isinstance(item, str):
            return self._append(GetItemOrAttribute(item_or_attribute=item))
        if isinstance(item, slice):
            return self._append(
                GetSlice(start=item.start, stop=item.stop, step=item.step)
            )
        if isinstance(item, Sequence):
            item = tuple(item)
            if all(isinstance(i, int) for i in item):
                return self._append(GetIndices(indices=item))
            elif all(isinstance(i, str) for i in item):
                return self._append(GetItems(items=item))
            else:
                raise TypeError(
                    f"Unhandled sequence item types: {list(map(type, item))}. "
                    f"Note mixing int and str is not allowed."
                )

        raise TypeError(f"Unhandled item type {type(item)}.")

    def __getattr__(self, attr: str) -> Lens:
        if attr == "_ipython_canary_method_should_not_exist_":
            # NOTE(piotrm): when displaying objects, ipython checks whether they
            # have overwritten __getattr__ by looking up this attribute. If it
            # does not result in AttributeError or None, IPython knows it was
            # overwritten and it will not try to use any of the _repr_*_ methods
            # to display the object. In our case, this will result Lenses being
            # constructed with this canary attribute name. We instead return
            # None here to let ipython know we have overwritten __getattr__ but
            # we do not construct any Lenses.
            return 0xDEAD

        return self._append(GetItemOrAttribute(item_or_attribute=attr))


Lens.model_rebuild()


class LensedDict(dict, Generic[T]):
    """A dictionary which can be accessed using lenses."""

    def __setitem__(self, __name: Union[str, Lens], __value: T) -> None:
        """Allow setitem to work on Lenses instead of just strings. Uses `Lens.set`
        if a lens is given."""

        if isinstance(__name, Lens):
            # Does not mutate so need to use dict.update .
            temp = __name.set(self, __value)
            self.update(temp)
            return None

        return super().__setitem__(__name, __value)

    def __getitem__(self, __name: Union[Any, Lens]) -> T:
        if isinstance(__name, Lens):
            return __name.get_sole_item(self)

        return super().__getitem__(__name)


# TODO: Deprecate old name.
JSONPath = Lens


def leaf_queries(obj_json: JSON, query: Lens = None) -> Iterable[Lens]:
    """
    Get all queries for the given object that select all of its leaf values.
    """

    query = query or Lens()

    if isinstance(obj_json, JSON_BASES):
        yield query

    elif isinstance(obj_json, Dict):
        for k, v in obj_json.items():
            sub_query = query[k]
            yield from leaf_queries(obj_json[k], sub_query)

    elif isinstance(obj_json, Sequence):
        for i, v in enumerate(obj_json):
            sub_query = query[i]
            yield from leaf_queries(obj_json[i], sub_query)

    else:
        yield query


def all_queries(obj: Any, query: Lens = None) -> Iterable[Lens]:
    """
    Get all queries for the given object.
    """

    query = query or Lens()

    if isinstance(obj, JSON_BASES):
        yield query

    elif isinstance(obj, pydantic.BaseModel):
        yield query

        for k in type(obj).model_fields:
            v = getattr(obj, k)
            sub_query = query[k]
            yield from all_queries(v, sub_query)

    elif isinstance(obj, Dict):
        yield query

        for k, v in obj.items():
            sub_query = query[k]
            yield from all_queries(obj[k], sub_query)

    elif isinstance(obj, Sequence):
        yield query

        for i, v in enumerate(obj):
            sub_query = query[i]
            yield from all_queries(obj[i], sub_query)

    else:
        yield query


def all_objects(obj: Any, query: Lens = None) -> Iterable[Tuple[Lens, Any]]:
    """
    Get all queries for the given object.
    """

    query = query or Lens()

    yield (query, obj)

    if isinstance(obj, JSON_BASES):
        pass

    elif isinstance(obj, pydantic.BaseModel):
        for k in type(obj).model_fields:
            v = getattr(obj, k)
            sub_query = query[k]
            yield from all_objects(v, sub_query)

    elif isinstance(obj, Dict):
        for k, v in obj.items():
            sub_query = query[k]
            yield from all_objects(obj[k], sub_query)

    elif isinstance(obj, Sequence):
        for i, v in enumerate(obj):
            sub_query = query[i]
            yield from all_objects(obj[i], sub_query)

    elif isinstance(obj, Iterable):
        logger.debug(
            f"Cannot create query for Iterable types like {obj.__class__.__name__} at query {query}. Convert the iterable to a sequence first."
        )

    else:
        logger.debug(f"Unhandled object type {obj} {type(obj)}")


def leafs(obj: Any) -> Iterable[Tuple[str, Any]]:
    for q in leaf_queries(obj):
        path_str = str(q)
        val = q(obj)
        yield (path_str, val)


def matching_objects(obj: Any, match: Callable) -> Iterable[Tuple[Lens, Any]]:
    for q, val in all_objects(obj):
        if match(q, val):
            yield (q, val)


def matching_queries(obj: Any, match: Callable) -> Iterable[Lens]:
    for q, _ in matching_objects(obj, match=match):
        yield q


# HACK013:
SerialModel.model_rebuild()
SerialBytes.model_rebuild()
Lens.model_rebuild()
Step.model_rebuild()
Collect.model_rebuild()
StepItemOrAttribute.model_rebuild()
GetAttribute.model_rebuild()
GetIndex.model_rebuild()
GetItem.model_rebuild()
GetItemOrAttribute.model_rebuild()
GetSlice.model_rebuild()
GetIndices.model_rebuild()
GetItems.model_rebuild()



"""
This file contains utility functions specific to certain span types.
"""

from __future__ import annotations

from inspect import signature
import json
import logging
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)

from opentelemetry.baggage import get_baggage
from opentelemetry.context import Context
from opentelemetry.trace.span import Span
from opentelemetry.util.types import AttributeValue
from trulens.otel.semconv.trace import ResourceAttributes
from trulens.otel.semconv.trace import SpanAttributes

if TYPE_CHECKING:
    from trulens.core.app import App

logger = logging.getLogger(__name__)

"""
Type definitions
"""
Attributes = Optional[
    Union[
        Dict[str, Any],
        Callable[
            [
                Optional[Any],
                Optional[Exception],
                List[Any],
                Optional[Dict[str, Any]],
            ],
            Dict[str, Any],
        ],
    ]
]

"""
General utilites for all spans
"""


def _stringify_span_attribute(o: Any) -> Tuple[bool, str]:
    """Converts an object to a string.

    Args:
        o: object to convert to a string

    Returns:
        A tuple containing:
        - A boolean indicating whether the object was jsonified (as opposed to
          simple stringification)
        - The string representation of the object.
    """
    try:
        return True, json.dumps(o)
    except Exception:
        pass
    # Special handling for exceptions to preserve their message
    if isinstance(o, Exception):
        try:
            return False, str(o)
        except Exception:
            return False, f"<{type(o).__name__}>"
    # For everything else, try str() but fallback to type name
    # This maintains backward compatibility with tests while avoiding
    # issues with objects that have poor __str__ implementations
    try:
        return False, str(o)
    except Exception:
        # If str() fails, use type name instead
        try:
            return False, f"<{type(o).__name__}>"
        except Exception:
            return False, "<unstringifiable>"


def _convert_to_valid_span_attribute_type(val: Any) -> AttributeValue:
    if isinstance(val, (bool, int, float, str)):
        return val
    if isinstance(val, (list, tuple)):
        for curr_type in [bool, int, float, str]:
            if all([isinstance(curr, curr_type) for curr in val]):
                return val
        jsonifiable, j = _stringify_span_attribute(val)
        if jsonifiable:
            return j
        return [_stringify_span_attribute(curr)[1] for curr in val]
    return _stringify_span_attribute(val)[1]


def set_span_attribute_safely(
    span: Span,
    key: str,
    value: Any,
) -> None:
    if value is not None:
        span.set_attribute(key, _convert_to_valid_span_attribute_type(value))


def set_string_span_attribute_from_baggage(
    span: Span, key: str, context: Optional[Context] = None
) -> None:
    value = get_baggage(key, context)
    if value is not None:
        span.set_attribute(key, str(value))


def validate_attributes(attributes: Dict[str, Any]) -> Dict[str, Any]:
    """
    Utility function to validate span attributes based on the span type.
    """
    if not isinstance(attributes, dict) or any([
        not isinstance(key, str) for key in attributes.keys()
    ]):
        raise ValueError("Attributes must be a dictionary with string keys.")

    if SpanAttributes.SPAN_TYPE in attributes:
        raise ValueError("Span type should not be set in attributes.")

    # TODO: validate Span type attributes.
    return attributes


def set_general_span_attributes(
    span: Span,
    /,
    span_type: SpanAttributes.SpanType,
    context: Optional[Context] = None,
) -> None:
    span.set_attribute(SpanAttributes.SPAN_TYPE, span_type)

    set_string_span_attribute_from_baggage(
        span, ResourceAttributes.APP_NAME, context
    )
    set_string_span_attribute_from_baggage(
        span, ResourceAttributes.APP_VERSION, context
    )
    set_string_span_attribute_from_baggage(
        span, ResourceAttributes.APP_ID, context
    )
    set_string_span_attribute_from_baggage(
        span, SpanAttributes.RECORD_ID, context
    )
    set_string_span_attribute_from_baggage(
        span, SpanAttributes.EVAL.TARGET_RECORD_ID, context
    )
    set_string_span_attribute_from_baggage(
        span, SpanAttributes.EVAL.EVAL_ROOT_ID, context
    )
    set_string_span_attribute_from_baggage(
        span, SpanAttributes.EVAL.METRIC_NAME, context
    )
    set_string_span_attribute_from_baggage(
        span, SpanAttributes.RUN_NAME, context
    )
    set_string_span_attribute_from_baggage(
        span, SpanAttributes.INPUT_ID, context
    )
    set_string_span_attribute_from_baggage(
        span, SpanAttributes.INPUT_RECORDS_COUNT, context
    )


def set_function_call_attributes(
    span: Span,
    ret: Any,
    func_name: str,
    func_exception: Optional[Exception],
    all_kwargs: Dict[str, Any],
) -> None:
    set_span_attribute_safely(span, SpanAttributes.CALL.RETURN, ret)
    set_span_attribute_safely(span, SpanAttributes.CALL.FUNCTION, func_name)
    set_span_attribute_safely(span, SpanAttributes.CALL.ERROR, func_exception)
    for k, v in all_kwargs.items():
        set_span_attribute_safely(span, f"{SpanAttributes.CALL.KWARGS}.{k}", v)


def set_user_defined_attributes(
    span: Span,
    *,
    attributes: Dict[str, Any],
) -> None:
    final_attributes = validate_attributes(attributes)

    cost_attributes = {}
    for key, value in final_attributes.items():
        if key.startswith(f"{SpanAttributes.COST.base}."):
            cost_attributes[key] = value
        else:
            set_span_attribute_safely(span, key, value)
    if cost_attributes:
        attributes_so_far = dict(getattr(span, "attributes", {}))
        if attributes_so_far:
            for cost_field in [
                SpanAttributes.COST.COST,
                SpanAttributes.COST.NUM_TOKENS,
                SpanAttributes.COST.NUM_PROMPT_TOKENS,
                SpanAttributes.COST.NUM_COMPLETION_TOKENS,
            ]:
                cost_attributes[cost_field] = cost_attributes.get(
                    cost_field, 0
                ) + attributes_so_far.get(cost_field, 0)
            currency = attributes_so_far.get(SpanAttributes.COST.CURRENCY, None)
            model = attributes_so_far.get(SpanAttributes.COST.MODEL, None)
            if currency not in [
                None,
                cost_attributes.get(SpanAttributes.COST.CURRENCY),
            ]:
                cost_attributes[SpanAttributes.COST.CURRENCY] = "mixed"
            if model not in [
                None,
                cost_attributes.get(SpanAttributes.COST.MODEL),
            ]:
                cost_attributes[SpanAttributes.COST.MODEL] = "mixed"
            for k, v in cost_attributes.items():
                set_span_attribute_safely(span, k, v)


"""
RECORD_ROOT SPAN
"""


def set_record_root_span_attributes(
    span: Span,
    /,
    func: Callable,
    args: tuple,
    kwargs: dict,
    ret: Any,
    exception: Optional[Exception],
) -> None:
    tru_app: App = get_baggage("__trulens_app__")
    sig = signature(func)
    input_selector = get_baggage("__trulens_input_selector__")

    # Use custom input selector if provided, otherwise fall back to default
    if input_selector is not None and callable(input_selector):
        try:
            main_input_value = input_selector(args, kwargs)
        except Exception as e:
            logger.warning(
                f"Custom input selector failed: {e}. Falling back to default."
            )
            main_input_value = tru_app.main_input(
                func, sig, sig.bind_partial(*args, **kwargs)
            )
    else:
        main_input_value = tru_app.main_input(
            func, sig, sig.bind_partial(*args, **kwargs)
        )

    set_span_attribute_safely(
        span,
        SpanAttributes.RECORD_ROOT.INPUT,
        main_input_value,
    )
    ground_truth_output = get_baggage(
        SpanAttributes.RECORD_ROOT.GROUND_TRUTH_OUTPUT
    )
    if ground_truth_output:
        set_span_attribute_safely(
            span,
            SpanAttributes.RECORD_ROOT.GROUND_TRUTH_OUTPUT,
            ground_truth_output,
        )
    if exception:
        set_span_attribute_safely(
            span, SpanAttributes.RECORD_ROOT.ERROR, str(exception)
        )
    if ret is not None:
        set_span_attribute_safely(
            span,
            SpanAttributes.RECORD_ROOT.OUTPUT,
            tru_app.main_output(
                func, sig, sig.bind_partial(*args, **kwargs), ret
            ),
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Any, List

import pandas as pd
from trulens.core.session import TruSession
from trulens.otel.semconv.trace import SpanAttributes

if TYPE_CHECKING:
    from trulens.core.app import App


class Recording:
    records: List[Record]

    def __init__(self, tru_app: App) -> None:
        self._tru_app = tru_app
        self.records = []

    def add_record_id(self, record_id: str) -> None:
        self.records.append(Record(self._tru_app, record_id))

    def get(self) -> Record:
        """
        Assumes there is exactly one record ID in the recording and returns it.

        Returns:
            The single record ID of the recording.
        """
        if len(self.records) == 0:
            raise RuntimeError("No records found!")
        if len(self.records) != 1:
            raise RuntimeError("There are multiple records!")
        self.records[0]._wait_for_record()
        return self.records[0]

    def __getitem__(self, index: int) -> Record:
        self.records[index]._wait_for_record()
        return self.records[index]

    def __len__(self) -> int:
        return len(self.records)

    def retrieve_feedback_results(self, timeout: float = 180) -> pd.DataFrame:
        """Retrieve feedback results for all records in the recording.

        Args:
            timeout: Timeout in seconds to wait.

        Returns:
            A dataframe with records as rows and feedbacks as columns.
        """
        return self._tru_app.retrieve_feedback_results(
            record_ids=[curr.record_id for curr in self.records],
            timeout=timeout,
        )

    # TODO(otel): record_metadata maybe?


# TODO(otel):
# This largely duplicates the functionality of
# `trulens.core.schema.record.Record` which was used for pre-OTel. Eventually,
# we should move to combining the two.
# Other functionality that needs to be added:
# 1. self.feedback_results
# 2. w/e get_feedback_result needs
# 3. w/e trulens_trace needs
# 4. w/e trulens_feedback needs
# 5. w/e display needs
# 6. w/e langchain_quickstart is doing.
# 7. w/e TruSession::run_feedback_functions needs.
class Record:
    """A class to represent a record."""

    def __init__(self, tru_app: App, record_id: str) -> None:
        self._tru_app = tru_app
        self._in_database = False
        self._events = None
        self._record_root_event = None
        self.record_id = record_id

    @property
    def main_input(self) -> Any:
        self._wait_for_record()
        record_root_event = self._get_record_root_event()
        return record_root_event["record_attributes"].get(
            SpanAttributes.RECORD_ROOT.INPUT
        )

    @property
    def main_output(self) -> Any:
        self._wait_for_record()
        record_root_event = self._get_record_root_event()
        return record_root_event["record_attributes"].get(
            SpanAttributes.RECORD_ROOT.OUTPUT
        )

    @property
    def main_error(self) -> Any:
        self._wait_for_record()
        record_root_event = self._get_record_root_event()
        return record_root_event["record_attributes"].get(
            SpanAttributes.RECORD_ROOT.ERROR
        )

    @property
    def meta(self) -> Any:
        # TODO(otel): Currently there's no metadata in OTel records!
        return None

    def _wait_for_record(self) -> None:
        if not self._in_database:
            TruSession().wait_for_records(
                record_ids=[self.record_id], timeout=180
            )
            self._in_database = True

    def _get_events(self, refresh: bool = False) -> pd.DataFrame:
        self._wait_for_record()
        if self._events is not None and not refresh:
            return self._events
        self._events = TruSession().connector.get_events(
            record_ids=[self.record_id]
        )
        if len(self._events) == 0:
            raise RuntimeError(
                f"No events found for record ID {self.record_id}!"
            )
        return self._events

    def _get_record_root_event(self, refresh: bool = False) -> pd.Series:
        self._wait_for_record()
        if self._record_root_event is not None and not refresh:
            return self._record_root_event
        events = self._get_events(refresh=refresh)
        record_root_events = [
            curr
            for _, curr in events.iterrows()
            if curr["record_attributes"].get(SpanAttributes.SPAN_TYPE)
            == SpanAttributes.SpanType.RECORD_ROOT
        ]
        if len(record_root_events) == 0:
            raise RuntimeError(
                f"No record root event found for record ID {self.record_id}!"
            )
        if len(record_root_events) > 1:
            raise RuntimeError(
                f"Multiple record root events found for record ID {self.record_id}!"
            )
        self._record_root_event = record_root_events[0]
        return self._record_root_event

    def retrieve_feedback_results(self, timeout: float = 180) -> pd.DataFrame:
        """Retrieve feedback results for the record.

        Args:
            timeout: Timeout in seconds to wait.

        Returns:
            A dataframe with a single row and feedbacks as columns.
        """
        return self._tru_app.retrieve_feedback_results(
            record_ids=[self.record_id], timeout=timeout
        )



"""create event table

Revision ID: 10
Revises: 9
Create Date: 2024-12-11 09:32:48.976169
"""

from alembic import op
import sqlalchemy as sa
from trulens.core.otel.utils import is_otel_tracing_enabled

# revision identifiers, used by Alembic.
revision = "10"
down_revision = "9"
branch_labels = None
depends_on = None


def _use_event_table():
    # We only use event table if specifically enabled as it requires the often
    # unsupported JSON type and is for temporary testing purposes anyway.
    return is_otel_tracing_enabled()


def upgrade(config) -> None:
    if not _use_event_table():
        return

    prefix = config.get_main_option("trulens.table_prefix")

    if prefix is None:
        raise RuntimeError("trulens.table_prefix is not set")

    # Check if the database is Snowflake
    dict_type = sa.JSON()
    is_snowflake = config.get_main_option("sqlalchemy.url", "").startswith(
        "snowflake"
    )
    if is_snowflake:
        # Snowflake does not support JSON type, so we use TEXT instead
        # for compatibility with the rest of the code.
        dict_type = sa.TEXT()

    op.create_table(
        prefix + "events",
        sa.Column("event_id", sa.VARCHAR(length=256), nullable=False),
        sa.Column("record", dict_type, nullable=False),
        sa.Column("record_attributes", dict_type, nullable=False),
        sa.Column("record_type", sa.VARCHAR(length=256), nullable=False),
        sa.Column("resource_attributes", dict_type, nullable=False),
        sa.Column("start_timestamp", sa.TIMESTAMP(), nullable=False),
        sa.Column("timestamp", sa.TIMESTAMP(), nullable=False),
        sa.Column("trace", dict_type, nullable=False),
        sa.PrimaryKeyConstraint("event_id"),
    )


def downgrade(config) -> None:
    if not _use_event_table():
        return

    prefix = config.get_main_option("trulens.table_prefix")

    if prefix is None:
        raise RuntimeError("trulens.table_prefix is not set")

    op.drop_table(prefix + "events")



"""
Utilities for app components provided as part of the trulens package.
Currently organizes all such components as "Other".
"""

from trulens.core import app
from trulens.core._utils.pycompat import Type
from trulens.core.utils import pyschema as pyschema_utils


class Other(app.Other, app.TrulensComponent):
    pass


# All component types, keep Other as the last one since it always matches.
COMPONENT_VIEWS = [Other]


def constructor_of_class(
    cls_obj: pyschema_utils.Class,
) -> Type[app.TrulensComponent]:
    for view in COMPONENT_VIEWS:
        if view.class_is(cls_obj):
            return view

    raise TypeError(f"Unknown trulens component type with class {cls_obj}")


def component_of_json(json: dict) -> app.TrulensComponent:
    cls = pyschema_utils.Class.of_class_info(json)

    view = constructor_of_class(cls)

    return view(json)



from typing import Union
import uuid

from opentelemetry import trace
from opentelemetry.baggage import get_baggage
from opentelemetry.baggage import remove_baggage
from opentelemetry.baggage import set_baggage
import opentelemetry.context as context_api
from opentelemetry.trace import get_current_span
from opentelemetry.trace.span import Span
from trulens.experimental.otel_tracing.core.session import TRULENS_SERVICE_NAME
from trulens.otel.semconv.trace import SpanAttributes


class UseCurrentSpanFunctionCallContextManager:
    def __enter__(self):
        return get_current_span()

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class CreateSpanFunctionCallContextManager:
    def __init__(self, span_name: str) -> None:
        self.span_name = span_name
        self.span_context_manager = None
        self.token = None
        self._started_record = False

    def __enter__(self) -> Span:
        # Set record_id into context.
        started_record = False
        record_id = get_baggage(SpanAttributes.RECORD_ID)
        if not record_id:
            started_record = True
            record_id = str(uuid.uuid4())
            self.token = context_api.attach(
                set_baggage(SpanAttributes.RECORD_ID, record_id)
            )
            recording = get_baggage("__trulens_recording__")
            if recording is not None:
                recording.add_record_id(record_id)
        self._started_record = started_record
        # Create span.
        self.span_context_manager = (
            trace.get_tracer_provider()
            .get_tracer(TRULENS_SERVICE_NAME)
            .start_as_current_span(name=self.span_name)
        )
        span = self.span_context_manager.__enter__()
        if started_record:
            span.set_attribute(
                SpanAttributes.SPAN_TYPE, SpanAttributes.SpanType.RECORD_ROOT
            )
        return span

    def __exit__(self, exc_type, exc_val, exc_tb):
        e_ret = None
        # Clean up span.
        try:
            if self.span_context_manager is not None:
                self.span_context_manager.__exit__(exc_type, exc_val, exc_tb)
                self.span_context_manager = None
        except Exception as e_span:
            e_ret = e_span
        # Clean up context.
        try:
            if self.token is not None:
                remove_baggage(SpanAttributes.RECORD_ID)
                context_api.detach(self.token)
                self.token = None
        except Exception as e_context:
            e_ret = e_context if e_ret is None else e_ret
        # Throw any errors we found.
        if e_ret is not None:
            raise e_ret


def create_function_call_context_manager(
    create_new_span: bool, span_name: str
) -> Union[
    UseCurrentSpanFunctionCallContextManager,
    CreateSpanFunctionCallContextManager,
]:
    if create_new_span:
        return CreateSpanFunctionCallContextManager(span_name)
    return UseCurrentSpanFunctionCallContextManager()
